# **************************************************************************
# *
# * Authors:    David Maluenda Niubo (dmaluenda@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import json
import subprocess

from pyworkflow.tests import *
import pyworkflow.utils as pwutils

from pwem import Domain
import pwem.protocols as emprot

# --- Set this to match with your queue system ---
#  json params to fill the queue form, see SCIPION_HOME/config/host.conf
QUEUE_PARAMS = (u'myslurmqueue', {u'JOB_TIME': u'1',        # in hours
                                  u'JOB_MEMORY': u'2048',   # in Mb
                                  u'QUEUE_FOR_JOBS': u'N', })
#  command and args to list the queued jobs (to be used in a subprocess)
#  the command's output must contain the jobID and the protocolID
QUEUE_COMMAND = ["squeue"]


class TestQueueBase(BaseTest):
    @classmethod
    def setData(cls, dataProject='mda'):
        cls.dataset = DataSet.getDataSet(dataProject)
        cls.particlesFn = cls.dataset.getFile('particles')

    @classmethod
    def runImportParticles(cls, pattern, samplingRate, checkStack=False):
        """ Run an Import particles protocol. """
        protImport = cls.newProtocol(emprot.ProtImportParticles, 
                                     filesPath=pattern,
                                     samplingRate=samplingRate,
                                     checkStack=checkStack)
        cls.launchProtocol(protImport)
        # check that input images have been imported (a better way to do this?)
        if protImport.outputParticles is None:
            raise Exception('Import of images: %s, failed. outputParticles '
                            'is None.' % pattern)
        return protImport

    @classmethod
    def runNormalizeParticles(cls, particles):
        """ Run normalize particles protocol """
        relionProtocols = Domain.importFromPlugin('relion.protocols',
                                                  doRaise=True)
        protPreproc = cls.newProtocol(relionProtocols.ProtRelionPreprocessParticles,
                                      doNormalize=True)
        protPreproc.inputParticles.set(particles)
        cls.launchProtocol(protPreproc)
        cls.sampling = protPreproc.outputParticles.getSamplingRate()
        return protPreproc

    def _checkAsserts(self, prot):
        """ If prot.useQueue() is True, first of all
            we check if the job is queued and after that
            we wait until it's finished to check the outputs
        """
        def isJobInQueue(jobId, protId, Yes=True):
            """ returns Yes if the protId and the jobId
                are found in the queue.
                Set Yes=False to get a True when the job disappears
            """
            isThere = not Yes
            queueRaw = subprocess.check_output(QUEUE_COMMAND)
            queueList = queueRaw.split('\n')
            for queueLine in queueList:
                # we will assume that if protId and jobId
                # is in the string, the task is there
                if jobId in queueLine and str(protId) in queueLine:
                    isThere = Yes
                    break
            return isThere

        def wait_until(condition, timeout, *args, **kwargs):
            """ :param condition: function handle that
                                      is waited until return True
                :param timeout: maximum time to wait
                :param args and kwargs: params to pass to the condition func.
                :return: False if the timeout is reached
            """
            mustend = time.time() + timeout
            while time.time() < mustend:
                if condition(*args, **kwargs):
                    return True
                time.sleep(1)
            return False

        def checkQueue(jobId, protId):
            """ Check if the protocol job is queued
            """
            self.assertTrue(isJobInQueue(jobId, protId),
                            "The job %s corresponding to "
                            "the protocol %d has been not "
                            "attached to the system queue"
                            % (jobId, protId))
            print(pwutils.magentaStr(" > job %s of the protocol %d found in the "
                                     "queue, wait a sec..." % (jobId, protId)))

            isDone = wait_until(isJobInQueue, 10*60, jobId, protId, Yes=False)
            self.assertTrue(isDone, "Timeout: the job has not ended...")
            print(pwutils.magentaStr("    ...job ended!"))

        if prot.useQueue():
            if QUEUE_PARAMS[1]['QUEUE_FOR_JOBS'] != 'Y':
                # if the protocol is use queue system, we check if it's queued
                jobId = prot.getJobId()   # is an string
                protId = prot.getObjId()  # is an integer
                checkQueue(jobId, protId)
                return  # I don't know why, but we cannot retrieve the output, permissions???
            else:
                # Check that job files have been created
                jobFilesPath = join(pwutils.getParentFolder(prot.getLogPaths()[0]),
                                    str(prot.getObjId()))

                self.assertTrue(
                    pwutils.exists(jobFilesPath + "-0-1.out") and pwutils.exists(
                        jobFilesPath + "-0-1.err") and pwutils.exists(jobFilesPath + "-0-1.job"),
                    "Job queue files not found in log folder, job did not make it to the queue.")

        self.assertIsNotNone(prot.outputClasses,
                             "There was a problem with Relion 2D classify")

        classsesPixSize = prot.outputClasses.getImages().getSamplingRate()
        self.assertAlmostEquals(self.sampling, classsesPixSize, delta=0.001,
                                msg="There was a problem with the sampling rate "
                                    "of the particles")
        for class2D in prot.outputClasses:
            self.assertTrue(class2D.hasAlignment2D())

    def _runRelionClassify2D(self, previousRun, label='', threads=1, MPI=1,
                             doGpu=False, GPUs='', useQueue=False, steps=False):
        """ :param previousRun: The outputParticles of that will be the input
            :param label: For naming purposes
            :param threads: How many threads to use
            :param MPI: How many MPIs to use
            :param doGpu: Use GPU or not
            :param GPUs: Which GPUs to use (see Relion gpusToUse form param)
            :param useQueue: Use the queue system or not
            :return: the launched protocol
        """
        relionProtocols = Domain.importFromPlugin('relion.protocols',
                                                  doRaise=True)
        prot2D = self.newProtocol(relionProtocols.ProtRelionClassify2D,
                                  doCTF=False, maskDiameterA=340,
                                  useGradientAlg=False,
                                  numberOfMpi=MPI, numberOfThreads=threads)
        prot2D.numberOfClasses.set(4)
        prot2D.numberOfIterations.set(3)
        prot2D.inputParticles.set(previousRun.outputParticles)
        prot2D.setObjLabel(label)

        if useQueue:
            prot2D._useQueue.set(True)
            if steps:
                QUEUE_PARAMS[1]['QUEUE_FOR_JOBS'] = 'Y'
            prot2D._queueParams.set(json.dumps(QUEUE_PARAMS))

        prot2D.doGpu.set(doGpu)
        if doGpu:
            prot2D.gpusToUse.set(GPUs)

        self.launchProtocol(prot2D)
        return prot2D


class TestQueueALL(TestQueueBase):
    @classmethod
    def setUpClass(cls):
        setupTestProject(cls)
        TestQueueBase.setData('mda')
        cls.protImport = cls.runImportParticles(cls.particlesFn, 3.5)
        cls.protNormalize = cls.runNormalizeParticles(cls.protImport.outputParticles)

    def testNoGpuSerial(self):
        relionNoGpu11 = self._runRelionClassify2D(self.protNormalize,
                                                  "Rel.2D noGPU serial",
                                                  useQueue=True)
        self._checkAsserts(relionNoGpu11)

    def testNoGpuMPI(self):
        relionNoGpu14 = self._runRelionClassify2D(self.protNormalize,
                                                  "Rel.2D noGPU MPI",
                                                  MPI=4, useQueue=True)
        self._checkAsserts(relionNoGpu14)

    def testNoGpuThreads(self):
        relionNoGpu41 = self._runRelionClassify2D(self.protNormalize,
                                                  "Rel.2D noGPU Threads",
                                                  threads=4, useQueue=True)
        self._checkAsserts(relionNoGpu41)

    def testNoGpuMPIandThreads(self):
        relionNoGpu22 = self._runRelionClassify2D(self.protNormalize,
                                                  "Rel.2D noGPU MPI+Threads",
                                                  MPI=2, threads=2,
                                                  useQueue=True)
        self._checkAsserts(relionNoGpu22)

    def testGpuSerial(self):
        relionGpu11 = self._runRelionClassify2D(self.protNormalize,
                                                "Rel.2D GPU serial",
                                                doGpu=True, useQueue=True)
        self._checkAsserts(relionGpu11)

    def testGpuMPI(self):
        relionGpu12 = self._runRelionClassify2D(self.protNormalize,
                                                "Rel.2D GPU MPI",
                                                doGpu=True, MPI=2,
                                                useQueue=True)
        self._checkAsserts(relionGpu12)

    def testGpuThreads(self):
        relionGpu41 = self._runRelionClassify2D(self.protNormalize,
                                                "Rel.2D GPU Threads",
                                                doGpu=True, threads=4,
                                                useQueue=True)
        self._checkAsserts(relionGpu41)

    def testGpuMPIandThreads(self):
        relionGpu22 = self._runRelionClassify2D(self.protNormalize,
                                                "Rel.2D GPU MPI+Threads",
                                                doGpu=True, useQueue=True,
                                                MPI=2, threads=2)
        self._checkAsserts(relionGpu22)


class TestQueueSmall(TestQueueBase):
    @classmethod
    def setUpClass(cls):
        setupTestProject(cls)
        TestQueueBase.setData('mda')
        cls.protImport = cls.runImportParticles(cls.particlesFn, 3.5)
        cls.protNormalize = cls.runNormalizeParticles(cls.protImport.outputParticles)

    def testGpuMPI(self):
        relionGpu12 = self._runRelionClassify2D(self.protNormalize,
                                                "Rel.2D GPU MPI",
                                                doGpu=True, MPI=2,
                                                useQueue=True)
        self._checkAsserts(relionGpu12)

    def testGpuMPISteps(self):
        relionGpu12 = self._runRelionClassify2D(self.protNormalize,
                                                "Rel.2D GPU MPI Steps",
                                                doGpu=True, MPI=4,
                                                useQueue=True, steps=True)

        self._checkAsserts(relionGpu12)


class TestNoQueueALL(TestQueueBase):
    @classmethod
    def setUpClass(cls):
        setupTestProject(cls)
        TestQueueBase.setData('mda')
        cls.protImport = cls.runImportParticles(cls.particlesFn, 3.5)
        cls.protNormalize = cls.runNormalizeParticles(cls.protImport.outputParticles)

    def testNoGpuSerial(self):
        relionNoGpu11 = self._runRelionClassify2D(self.protNormalize,
                                                  "Rel.2D noGPU serial")
        self._checkAsserts(relionNoGpu11)

    def testNoGpuMPI(self):
        relionNoGpu14 = self._runRelionClassify2D(self.protNormalize,
                                                  "Rel.2D noGPU MPI",
                                                  MPI=4)
        self._checkAsserts(relionNoGpu14)

    def testNoGpuThreads(self):
        relionNoGpu41 = self._runRelionClassify2D(self.protNormalize,
                                                  "Rel.2D noGPU Threads",
                                                  threads=4)
        self._checkAsserts(relionNoGpu41)

    def testNoGpuMPIandThreads(self):
        relionNoGpu22 = self._runRelionClassify2D(self.protNormalize,
                                                  "Rel.2D noGPU MPI+Threads",
                                                  MPI=2, threads=4)
        self._checkAsserts(relionNoGpu22)

    def testGpuSerial(self):
        relionGpu11 = self._runRelionClassify2D(self.protNormalize,
                                                "Rel.2D GPU serial",
                                                doGpu=True)
        self._checkAsserts(relionGpu11)

    def testGpuMPI(self):
        relionGpu12 = self._runRelionClassify2D(self.protNormalize,
                                                "Rel.2D GPU MPI",
                                                doGpu=True, MPI=2)
        self._checkAsserts(relionGpu12)

    def testGpuThreads(self):
        relionGpu41 = self._runRelionClassify2D(self.protNormalize,
                                                "Rel.2D GPU Threads",
                                                doGpu=True, threads=4)
        self._checkAsserts(relionGpu41)

    def testGpuMPIandThreads(self):
        relionGpu22 = self._runRelionClassify2D(self.protNormalize,
                                                "Rel.2D GPU MPI+Threads",
                                                doGpu=True,
                                                MPI=2, threads=4)
        self._checkAsserts(relionGpu22)


class TestNoQueueSmall(TestQueueBase):
    @classmethod
    def setUpClass(cls):
        setupTestProject(cls)
        TestQueueBase.setData('mda')
        cls.protImport = cls.runImportParticles(cls.particlesFn, 3.5)
        cls.protNormalize = cls.runNormalizeParticles(cls.protImport.outputParticles)

    def testGpuMPI(self):
        relionGpu12 = self._runRelionClassify2D(self.protNormalize,
                                                "Rel.2D GPU MPI",
                                                doGpu=True, MPI=2)
        self._checkAsserts(relionGpu12)


class TestQueueSteps(TestQueueBase):
    @classmethod
    def setUpClass(cls):
        setupTestProject(cls)
        TestQueueBase.setData('mda')
        cls.protImport = cls.runImportParticles(cls.particlesFn, 3.5)

    def testStepsNoGPU(self):
        xmipp3Protocols = Domain.importFromPlugin('xmipp3.protocols',
                                                  doRaise=True)
        protXmippPreproc = self.newProtocol(xmipp3Protocols.XmippProtPreprocessParticles,
                                            doNormalize=True, doRemoveDust=True)

        protXmippPreproc.inputParticles.set(self.protImport.outputParticles)
        protXmippPreproc.setObjLabel("Xmipp preprocess steps")

        protXmippPreproc._useQueue.set(True)

        QUEUE_PARAMS[1]['QUEUE_FOR_JOBS'] = 'Y'

        protXmippPreproc._queueParams.set(json.dumps(QUEUE_PARAMS))

        # Launch protocol but wait until it finishes
        self.launchProtocol(protXmippPreproc, wait=True)

        # Check that job files have been created

        jobFilesPath = join(pwutils.getParentFolder(protXmippPreproc.getLogPaths()[0]),
                            str(protXmippPreproc.getObjId()))

        self.assertTrue(pwutils.exists(jobFilesPath + "-0-1.out") and
                        pwutils.exists(jobFilesPath + "-0-1.err") and
                        pwutils.exists(jobFilesPath + "-0-1.job") and
                        pwutils.exists(jobFilesPath + "-0-2.out") and
                        pwutils.exists(jobFilesPath + "-0-2.err") and
                        pwutils.exists(jobFilesPath + "-0-2.job"),
                        "Job queue files not found on log folder, job did "
                        "not make it to the queue.")

        # Check that results have been produced
        self.assertIsNotNone(protXmippPreproc.outputParticles,
                             "There was a problem with Xmipp preprocess particles.")




# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import csv
import os
import re

from pwem.objects import SetOfClasses2D
from pyworkflow.config import Config as pwConfig
from pyworkflow.gui import getDefaultFont, openTextFileEditor

import tkinter as tk
import tkinter.ttk as ttk

import pyworkflow.viewer as pwviewer

from pwem import emlib

from .showj import (getJvmMaxMemory, MODE, VISIBLE, ZOOM, ORDER, RENDER,
                    SORT_BY, runJavaIJapp, MODE_MD, LABELS,
                    launchSupervisedPickerGUI)


class DataView(pwviewer.View):
    """ Wrapper the arguments to showj (either web or desktop).
    Also useful to visualize images that are not objects,
    e.g.: dark or gain images"""

    def __init__(self, path, viewParams={}, **kwargs):
        pwviewer.View.__init__(self)
        self._memory = getJvmMaxMemory()
        self._loadPath(path)
        self._env = kwargs.get('env', {})
        self._viewParams = viewParams

    def setMemory(self, memory):
        self._memory = memory

    def getViewParams(self):
        """ Give access to the viewParams dict. """
        return self._viewParams

    def _loadPath(self, path):
        self._tableName = None

        # If path is a tuple, we will convert to the filename format
        # as expected by Showj
        if isinstance(path, tuple):
            self._path = emlib.image.ImageHandler.locationToXmipp(path)
        # Check if there is a table name with @ in path
        # in that case split table name and path
        # table names can never starts with a number
        # this is considering an image inside an stack
        elif isinstance(path, str):
            if '@' in path and path[0] not in '0123456789':
                self._tableName, self._path = path.split('@')
            else:
                self._path = path
        else:
            raise Exception("Invalid input path, "
                            "should be 'string' or 'tuple'")

    def show(self):
        runJavaIJapp(self._memory,
                     'xmipp.viewer.scipion.ScipionViewer',
                     self.getShowJParams(), env=self._env)

    def getShowJParams(self):
        tableName = '%s@' % self._tableName if self._tableName else ''
        params = '-i "%s%s"' % (tableName, self._path)

        for key, value in self._viewParams.items():
            params = "%s --%s %s" % (params, key, value)

        if pwConfig.debugOn():
            params += " --debug"

        return params

    def getShowJWebParams(self):

        parameters = {
            MODE,  # FOR MODE TABLE OR GALLERY
            VISIBLE,
            ZOOM,
            ORDER,
            RENDER,
            SORT_BY
        }

        params = {}

        for key, value in self._viewParams.items():
            if key in parameters:
                if key == 'mode' and value == 'metadata':
                    value = 'table'
                params[key] = value

        return params

    def getPath(self):
        return self._path

    def getTableName(self):
        return self._tableName


class ObjectView(DataView):
    """ Wrapper to DataView but for displaying Scipion objects. """

    def __init__(self, project, inputid, path, other='', viewParams={},
                 **kwargs):
        DataView.__init__(self, path, viewParams, **kwargs)
        self.type = type
        self.port = project.port
        self.inputid = inputid
        self.other = other

    def getShowJParams(self):
        # Add the scipion parameters over the normal showj params
        return '%s --scipion %s %s %s' % (DataView.getShowJParams(self),
                                          self.port, self.inputid, self.other)

    def show(self):
        runJavaIJapp(self._memory, 'xmipp.viewer.scipion.ScipionViewer',
                     self.getShowJParams(), env=self._env)


class MicrographsView(ObjectView):
    """ Customized ObjectView for SetOfCTF objects . """
    # All extra labels that we want to show if present in the CTF results
    RENDER_LABELS = ['thumbnail._filename', 'psdCorr._filename',
                     'plotGlobal._filename']
    EXTRA_LABELS = ['_filename']

    def __init__(self, project, micSet, other='', **kwargs):
        first = micSet.getFirstItem()

        def existingLabels(labelList):
            return ' '.join([l for l in labelList if first.hasAttributeExt(l)])

        renderLabels = existingLabels(self.RENDER_LABELS)
        extraLabels = existingLabels(self.EXTRA_LABELS)
        labels = 'id enabled %s %s' % (renderLabels, extraLabels)

        viewParams = {MODE: MODE_MD,
                      ORDER: labels,
                      VISIBLE: labels,
                      ZOOM: 50
                      }

        if renderLabels:
            viewParams[RENDER] = renderLabels

        inputId = micSet.getObjId() or micSet.getFileName()
        ObjectView.__init__(self, project,
                            inputId, micSet.getFileName(), other,
                            viewParams, **kwargs)


class CtfView(ObjectView):
    """ Customized ObjectView for SetOfCTF objects . """
    # All extra labels that we want to show if present in the CTF results
    PSD_LABELS = ['_micObj.thumbnail._filename', '_psdFile',
                  '_xmipp_enhanced_psd', '_xmipp_ctfmodel_quadrant',
                  '_xmipp_ctfmodel_halfplane', '_micObj.plotGlobal._filename'
                  ]
    EXTRA_LABELS = ['_ctftilt_tiltAxis', '_ctftilt_tiltAngle',
                    '_xmipp_ctfCritFirstZero',
                    '_xmipp_ctfCritCorr13', '_xmipp_ctfCritIceness', '_xmipp_ctfCritFitting',
                    '_xmipp_ctfCritNonAstigmaticValidty',
                    '_xmipp_ctfCritCtfMargin', '_xmipp_ctfCritMaxFreq',
                    '_xmipp_ctfCritPsdCorr90', '_xmipp_ctfVPPphaseshift'
                    ]

    def __init__(self, project, ctfSet, other='', **kwargs):
        first = ctfSet.getFirstItem()

        def existingLabels(labelList):
            return ' '.join([l for l in labelList if first.hasAttributeExt(l)])

        psdLabels = existingLabels(self.PSD_LABELS)
        extraLabels = existingLabels(self.EXTRA_LABELS)
        labels = 'id enabled %s _defocusU _defocusV ' % psdLabels
        labels += '_defocusAngle _defocusRatio '
        labels += '_phaseShift _resolution _fitQuality %s ' % extraLabels
        labels += ' _micObj._filename'

        viewParams = {MODE: MODE_MD,
                      ORDER: labels,
                      VISIBLE: labels,
                      ZOOM: 50
                      }

        if psdLabels:
            viewParams[RENDER] = psdLabels

        if ctfSet.isStreamOpen():
            viewParams['dont_recalc_ctf'] = ''

        inputId = ctfSet.getObjId() or ctfSet.getFileName()
        ObjectView.__init__(self, project,
                            inputId, ctfSet.getFileName(), other,
                            viewParams, **kwargs)


class ClassesView(ObjectView):
    """ Customized ObjectView for SetOfClasses. """

    def __init__(self, project, inputid, path, other='',
                 viewParams={}, **kwargs):
        from pwem.viewers.viewers_data import RegistryViewerConfig
        config = RegistryViewerConfig.getConfig(SetOfClasses2D)

        defaultViewParams = config.copy()

        defaultViewParams.update(viewParams)
        ObjectView.__init__(self, project, inputid, path, other,
                            defaultViewParams, **kwargs)


class Classes3DView(ClassesView):
    """ Customized ObjectView for SetOfClasses. """

    def __init__(self, project, inputid, path, other='',
                 viewParams={}, **kwargs):
        defaultViewParams = {ZOOM: '99',
                             MODE: 'metadata'}
        defaultViewParams.update(viewParams)
        ClassesView.__init__(self, project, inputid, path, other,
                             defaultViewParams, **kwargs)


class CoordinatesObjectView(DataView):
    """ Wrapper to View but for displaying Scipion objects. """
    MODE_AUTOMATIC = 'Automatic'

    def __init__(self, project, path, outputdir, protocol, pickerProps=None,
                 inTmpFolder=False, **kwargs):
        DataView.__init__(self, path, **kwargs)
        self.project = project
        self.outputdir = outputdir
        self.protocol = protocol
        self.pickerProps = pickerProps
        self.inTmpFolder = inTmpFolder
        self.mode = kwargs.get('mode', None)

    def show(self):
        return launchSupervisedPickerGUI(self._path, self.outputdir,
                                         self.protocol, mode=self.mode,
                                         pickerProps=self.pickerProps,
                                         inTmpFolder=self.inTmpFolder)


class ImageView(pwviewer.View):
    """ Customized ObjectView for SetOfClasses. """

    def __init__(self, imagePath, **kwargs):
        pwviewer.View.__init__(self)
        self._imagePath = os.path.abspath(imagePath)

    def getImagePath(self):
        return self._imagePath


class TableView(pwviewer.View):
    """ show table, pass values as:
        headerList = ['name', 'surname']
        dataList = [
        ('John', 'Smith') ,
        ('Larry', 'Black') ,
        ('Walter', 'White') ,
        ('Fred', 'Becker')
        ].
        msg = message to be shown at the table top
        title= window title
        height: Specifies the number of rows which should be visible
        width: minimum width in pixels
        fontSize= font size
        padding: cell extra width
        ---------------------
        Alternative way to create a table using showj
        views = []
        labels = '_1 _2'
        emSet = EMSet(filename="/tmp/kk.sqlite")
        emObject = EMObject()
        emObject._1 = String('first parameter')
        emObject._2 = Float(12.)
        emSet.append(emObject)
        emObject = EMObject()
        emObject._1 = String('second parameter')
        emObject._2 = Float(22.)
        emSet.append(emObject)
        emSet.write()
        views.append(ObjectView(self._project,
                                self.protocol.strId(),
                                "/tmp/kk.sqlite",
                                viewParams = {MODE: MODE_MD, ORDER: labels,
                                VISIBLE: labels}))
        return views
"""

    def __init__(self, headerList, dataList,
                 mesg=None, title=None,
                 height=10, width=400,
                 padding=10, outFileName=None):
        """Get new widget that has as parent the top level window and set title.
        tables can be dump to a csv file. The name of the file is outFileName
        """
        self.title = title
        self.headerList = headerList
        self.dataList = dataList
        self.outFileName = outFileName
        win = tk.Toplevel()
        if title:
            win.wm_title(title)

        # frame to place all other widgets
        frame = tk.Frame(win)

        font = getDefaultFont()
        font.metrics()
        fontheight = font.metrics()['linespace']
        style = ttk.Style()
        style.configure('Calendar.Treeview', font=font, rowheight=fontheight)

        # add a "save data" button
        saveDataButton = ttk.Button(master=win, text="Save Data",
                                    command=self.saveData)
        # create treeview to store multi list with data
        tree = ttk.Treeview(columns=headerList,
                            show="headings", master=win,
                            style='Calendar.Treeview', height=height
                            )

        # define scrollbars to be added
        if len(dataList) > height:
            ysb = ttk.Scrollbar(orient=tk.VERTICAL, command=tree.yview,
                                master=win)
            # xsb = ttk.Scrollbar(orient=tk.HORIZONTAL,
            # command= tree.xview, master=win)
            # add them to three view
            tree.configure(yscroll=ysb.set)  # , xscroll=xsb.set)

        # create rows and columns
        counterRow = 1
        colWidths = []  # list with maximum width per column
        # create headers
        for col in headerList:
            tree.heading(col, text=col.title())
            # save neede width for this cell
            (w, h) = (font.measure(col.title()), font.metrics("linespace"))
            colWidths.append(w)

        # insert other rows
        # tag rows as odd or even so they may have different background colors
        for item in dataList:
            if counterRow % 2:
                tree.insert('', 'end', values=item, tags=('evenrow',))
            else:
                tree.insert('', 'end', values=item, tags=('oddrow',))
            counterRow += 1
            counterCol = 0
            for i in item:
                (w, h) = (font.measure(i), font.metrics("linespace"))
                if colWidths[counterCol] < w:
                    colWidths[counterCol] = w
                counterCol += 1

        # if width less than sum of column widths expand them
        sumColWid = sum(colWidths) + 20
        if sumColWid < width:
            sumColWid = width
            factor = int(width / sumColWid) + 1
            colWidths = [i * factor for i in colWidths]

        for col, colWidth in zip(headerList, colWidths):
            tree.column(col, width=colWidth + padding)

        # color by rows
        tree.tag_configure('evenrow', background='white')
        tree.tag_configure('oddrow', background='light grey')

        # message placed at the window top
        msg = ttk.Label(wraplength=sumColWid, justify="left", anchor="n",
                        padding=(10, 2, 10, 6), text=mesg, master=win,
                        font=font)

        # set mg in grid 0,0
        msg.grid(row=0, column=0)
        # set button in grid 1 0
        saveDataButton.grid(row=1, column=0)
        # set ysg in grid 2 0
        tree.grid(row=2, column=0)
        # set ysg in grid 2 1
        # but only if number of elements is larger than height
        if len(dataList) > height:
            ysb.grid(row=2, column=1, sticky='ns')

    def saveData(self):
        """ Save data in the treeView Widget to a file.
        If not provided the outfile is in the Logs directory
        the name is the title of the table"""
        def slugify(value):
            """
            Create a valid file name from the table title.
            Normalizes string, converts to lowercase, removes non-alpha characters,
            and converts spaces to hyphens.
            """
            value = re.sub('[^\w\s-]', '', value).strip().lower()
            value = re.sub('[-\s]+', '-', value)
            return value

        if self.outFileName is None:
            outFileName = slugify(self.title) + ".csv"
            outFileName = os.path.join(os.getcwd(), 'Logs', outFileName)

        print("Data saved to file: ", outFileName)
        with open(outFileName, "wt") as f:
            f.write("# TITLE = %s\n" % self.title)
            spamWriter = csv.writer(f)
            spamWriter.writerow(self.headerList)
            for data in self.dataList:
                spamWriter.writerow(data)
        f.close()
        # call the application used to handle csv files.
        openTextFileEditor(outFileName)



# **************************************************************************
# *
# * Authors:     J.M. de la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from collections import OrderedDict

from pyworkflow.object import ObjectWrap

from ..lib import *


LABEL_TYPES = {
    LABEL_SIZET: float,
    LABEL_DOUBLE: float,
    LABEL_INT: int,
    LABEL_BOOL: bool
}


class Row:
    """ Support Xmipp class to store label and value pairs
    corresponding to a Metadata row.
    """

    def __init__(self):
        self._labelDict = OrderedDict()  # Dictionary containing labels and values
        self._objId = None  # Set this id when reading from a metadata

        # Add set and get method as alias to setValue and getValue
        # in this way the Row will behave more like a dict
        self.set = self.setValue
        self.get = self.getValue

    def getObjId(self):
        return self._objId

    def hasLabel(self, label):
        return self.containsLabel(label)

    def containsLabel(self, label):
        # Allow getValue using the label string
        if isinstance(label, str):
            label = str2Label(label)
        return label in self._labelDict

    def removeLabel(self, label):
        if self.hasLabel(label):
            del self._labelDict[label]

    def setValue(self, label, value):
        """args: this list should contains tuples with
        MetaData Label and the desired value"""
        # Allow setValue using the label string
        if isinstance(label, str):
            label = str2Label(label)
        self._labelDict[label] = value

    def getValue(self, label, default=None):
        """ Return the value of the row for a given label. """
        # Allow getValue using the label string
        if isinstance(label, str):
            label = str2Label(label)
        return self._labelDict.get(label, default)

    def getValueAsObject(self, label, default=None):
        """ Same as getValue, but making an Object wrapping. """
        return ObjectWrap(self.getValue(label, default))

    def readFromMd(self, md, objId):
        """ Get all row values from a given id of a metadata. """
        self._objId = objId
        try:
            self._labelDict = md.getRow(objId)
        except AttributeError:  # can be eventually removed, once we are sure that users have recent Xmipp
            self._labelDict.clear()
            for label in md.getActiveLabels():
                self._labelDict[label] = md.getValue(label, objId)

    def addToMd(self, md):
        self.writeToMd(md, md.addObject())

    def writeToMd(self, md, objId):
        """ Set back row values to a metadata row. """
        try:
            md.setRow(self._labelDict, objId)
        except AttributeError:  # can be eventually removed, once we are sure that users have recent Xmipp
            for label, value in self._labelDict.items():
                # TODO: Check how to handle correctly unicode type
                # in Xmipp and Scipion
                if type(value) is str:
                    value = str(value)
                try:
                    if labelType(label) == 2:
                        md.setValue(label, float(value), objId)
                    else:
                        md.setValue(label, value, objId)
                except Exception as ex:
                    import sys
                    print("XmippMdRow.writeToMd: Error writing value to metadata.", file=sys.stderr)
                    print("                     label: %s, value: %s, type(value): %s" % (
                        label2Str(label), value, type(value)), file=sys.stderr)
                    raise ex

    def readFromFile(self, fn):
        md = MetaData(fn)
        self.readFromMd(md, md.firstObject())

    def writeToFile(self, fn):
        md = MetaData()
        self.writeToMd(md, md.addObject())
        md.write(fn)

    def copyFromRow(self, other):
        for label, value in other._labelDict.items():
            self.setValue(label, value)

    def clone(self):
        """ Return another Row that have exactly the same
        values as self.
        """
        row = Row()
        row.copyFromRow(self)
        row._objId = self._objId

        return row

    def __str__(self):
        s = '{'
        for k, v in self._labelDict.items():
            s += '  %s = %s\n' % (label2Str(k), v)
        return s + '}'

    def __iter__(self):
        return self._labelDict.items()

    def __getitem__(self, item):
        return self.getValue(item)

    def __setitem__(self, key, value):
        return self.setValue(key, value)

    def containsAll(self, labels):
        """ Check if all labels are present in the row.
        Params:
            row: the Row object.
            labels: either a dict or list object containing the labels
                (in the case of dicts, label are the dict.values())
        """
        values = labels.values() if isinstance(labels, dict) else labels
        return all(self.containsLabel(l) for l in values)

    def containsAny(self, labels):
        """ Check if at least one of labels is present in the row.
        Params:
            row: the Row object.
            labels: either a dict or list object containing the labels
                (in the case of dicts, label are the dict.values())
        """
        values = labels.values() if isinstance(labels, dict) else labels
        return any(self.containsLabel(l) for l in values)

    def printDict(self):
        """ Fancy printing of the row, mainly for debugging. """
        print(str(self))


class RowMetaData:
    """ This class is a wrapper for MetaData in row mode.
    Where only one object is used.
    """

    def __init__(self, filename=None):
        self._md = MetaData()
        self._md.setColumnFormat(False)
        self._id = self._md.addObject()

        if filename:
            self.read(filename)

    def setValue(self, label, value):
        self._md.setValue(label, value, self._id)

    def getValue(self, label):
        return self._md.getValue(label, self._id)

    def write(self, filename, mode=MD_APPEND):
        self._md.write(filename, mode)

    def read(self, filename):
        self._md.read(filename)
        self._md.setColumnFormat(False)
        self._id = self._md.firstObject()

    def containsLabel(self, label):
        return self._md.containsLabel(label)

    def __str__(self):
        return str(self._md)


def label2Python(label):
    """ Return the Python type (int, float, bool) for a given 
    metadata label (LABEL_INT, LABEL_DOUBLE..etc)
    """
    return LABEL_TYPES.get(labelType(label), str)


def getLabel(value):
    """ Return the label value either from an int value or an string. """
    if isinstance(value, int):
        return value
    elif isinstance(value, str):
        return str2Label(value)
    else:
        raise Exception("Invalid value type (%s) for label. " % type(value))


def getFirstRow(mdOrFn):
    """ Return the first object of a metadata.
    Params:
        mdOrFn: you can pass a metadata or a filename as argument.
    """

    if isinstance(mdOrFn, str):
        md = MetaData()
        md.read(mdOrFn, 1)
    else:  # mdOrFn is MetaData
        md = mdOrFn
        
    if md.getParsedLines():
        row = Row()
        row.readFromMd(md, md.firstObject())
    else:
        row = None
    
    return row


def getSize(filename):
    """ Return the metadata size without parsing entirely. """
    md = MetaData()
    md.read(filename, 1)
    return md.getParsedLines()


def isEmpty(filename):
    """ Use getMdSize to check if metadata is empty. """
    return getSize(filename) == 0


def iterRows(md, sortByLabel=None):
    """ Iterate over the rows of the given metadata.
    Params:
        md: a MetaData object or a filename (MetaData will be read)
        sortByLabel: a label to sort the metadata before iterate.
    """
    # If md is string, take as filename and create the metadata

    if isinstance(md, str):
        md = MetaData(md)

    if sortByLabel is not None:
        md.sort(sortByLabel)

    row = Row()
    
    for objId in md:
        row.readFromMd(md, objId)
        yield row


def dropColumns(mdObj, *labels):
    """ Drop all columns from a given metadata.
    Labels can be either string or int.
    """
    for l in labels:
        mdObj.removeLabel(getLabel(l))


def keepColumns(mdObj, *labels):
    """ Drop all columns from mdObj that are not in labels.
    Labels can be either string or int.
    """
    # Handle string or int labels input
    keepLabels = {getLabel(l) for l in labels}

    for l in mdObj.getActiveLabels():
        if l not in keepLabels:
            mdObj.removeLabel(l)


def joinBlocks(inputMd, blockPrefix=None):
    mdImages = MetaData()
    mdAll = MetaData()
    mdBlocks = getBlocksInMetaDataFile(inputMd)
    
    for mdBlock in mdBlocks:
        if blockPrefix is not None:
            if mdBlock.startswith(blockPrefix):
                mdImages.read(mdBlock + "@" + inputMd)
                mdAll.unionAll(mdImages)
        else:
            mdImages.read(mdBlock + "@" + inputMd)
            mdAll.unionAll(mdImages)
    return mdAll


class SetMdIterator:
    """ Class to iterate over an input set and skip
    elements not present in metadata.
    This class can be used in copyItems when the number
    of elements in the set is higher that in metadata.
    """
    def __init__(self, md, sortByLabel=None, 
                 keyLabel=MDL_ITEM_ID,
                 updateItemCallback=None,
                 skipDisabled=False):
        
        if updateItemCallback is None:
            raise Exception('Set an updateItemCallback')
        
        self.iterMd = iterRows(md, sortByLabel) 
        self.keyLabel = keyLabel
        self.updateItemCallback = updateItemCallback
        self.skipDisabled = skipDisabled
        self.__nextRow()
        
    def __nextRow(self):
        try:
            self.lastRow = next(self.iterMd)
        except StopIteration:
            self.lastRow = None
            
    def updateItem(self, item, row):
        """ This function should be passed to copyItems
        as callback and it will filter the items
        not present in the metadata.
        """
        row = self.lastRow
        
        if row is not None:
            if row.hasLabel(MDL_ENABLED):
                enabled = row.getValue(MDL_ENABLED)
            else:
                enabled = 1

        if (row is None or
                item.getObjId() != row.getValue(self.keyLabel)):
            item._appendItem = False
            
        elif enabled == -1 and self.skipDisabled:
            item._appendItem = False
            self.__nextRow()
            
        else:
            item._appendItem = True
            self.updateItemCallback(item, row)
            self.__nextRow()



# **************************************************************************
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import pyworkflow.tests as pwtests

from pwem import Domain
import pwem.protocols as emprot

from .test_workflow import TestWorkflow


# update this test when RCT workflow are implemented
class TestXmippAssignmentTiltPairsWorkflow(TestWorkflow):
    @classmethod
    def setUpClass(cls):
        pwtests.setupTestProject(cls)
        cls.dataset = pwtests.DataSet.getDataSet('rct')
        cls.allCrdsDir = cls.dataset.getFile('positions')
        cls.micsUFn = cls.dataset.getFile('untilted')
        cls.micsTFn = cls.dataset.getFile('tilted')

    def test1(self):
        # First, import a set of micrographs
        protImport = self.newProtocol(emprot.ProtImportMicrographsTiltPairs,
                                      patternUntilted=self.micsUFn,
                                      patternTilted=self.micsTFn,
                                      samplingRate=2.28, voltage=100,
                                      sphericalAberration=2.9)
        self.launchProtocol(protImport)
        self.assertIsNotNone(protImport.outputMicrographsTiltPair,
                             "There was a problem with the import")

        protImportCoorU = self.newProtocol(emprot.ProtImportCoordinates,
                                           importFrom=emprot.ProtImportCoordinates.IMPORT_FROM_XMIPP,
                                           filesPath=self.allCrdsDir,
                                           filesPattern='F_rct_u_*.pos', boxSize=100)
        uMics = protImport.outputMicrographsTiltPair.getUntilted()
        protImportCoorU.inputMicrographs.set(uMics)
        self.launchProtocol(protImportCoorU)

        tMics = protImport.outputMicrographsTiltPair.getTilted()
        protImportCoorT = self.newProtocol(emprot.ProtImportCoordinates,
                                           importFrom=emprot.ProtImportCoordinates.IMPORT_FROM_XMIPP,
                                           filesPath=self.allCrdsDir,
                                           filesPattern='F_rct_t_*.pos', boxSize=100)
        protImportCoorT.inputMicrographs.set(tMics)
        self.launchProtocol(protImportCoorT)

        # Then simulate a particle picking
        print("Running tilt pairs assignment...")

        XmippProtAssignmentTiltPair = Domain.importFromPlugin(
            'xmipp3.protocols',
            'XmippProtAssignmentTiltPair',
            doRaise=True)
        protAssigning = self.newProtocol(XmippProtAssignmentTiltPair)
        micsTiltPair = protImport.outputMicrographsTiltPair
        protAssigning.inputMicrographsTiltedPair.set(micsTiltPair)
        print(self.micsUFn)
        print(self.micsTFn)
        protAssigning.untiltedSet.set(protImportCoorU.outputCoordinates)
        protAssigning.tiltedSet.set(protImportCoorT.outputCoordinates)
        self.launchProtocol(protAssigning)
        self.assertIsNotNone(protAssigning.outputCoordinatesTiltPair,
                             "There was a problem with the protocol assignment tilt pairs")
        print('-----------------------------------------------------------')
        num_particles = protAssigning.outputCoordinatesTiltPair.getUntilted().getSize()
        print(num_particles)
        if num_particles > 1000:
            out_ = True
        else:
            out_ = None
        self.assertIsNotNone(out_, "There was a problem with the protocol assignment tilt pairs")

        # self.validateFiles('protPicking', protPicking)



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This module contains protocols related to Set operations such us:
- subsets
- unions
- split
... etc
"""

import random
import sys

import pyworkflow.protocol as pwprot
import pyworkflow.object as pwobj

import pwem.objects as emobj
from pwem.protocols import EMProtocol
from pwem.objects import Volume, EMSet
from pyworkflow.utils import ProgressBar, getListFromRangeString


class ProtSets(EMProtocol):
    """ Base class for all protocols related to subsets. """

    def _append(self, outputSet, item, sourceItem=None):
        """ Add an item to the outputSet.
        If the item is a new copy of sourceItem(case of the join sets),
        then use a sourceItem since item lost the information related with the
        mapper
        """
        subElemList = []
        if sourceItem is None:
            sourceItem = item
        if isinstance(item, EMSet):
            for subElem in sourceItem.iterItems():
                # We need to create a clone because all items have a same _objId
                subElemList.append(subElem.clone())

        outputSet.append(item)
        if subElemList:
            for subElem in subElemList:
                item.append(subElem)


class ProtUnionSet(ProtSets):
    """ Protocol to join two or more sets of images.
    This protocol allows to select two or more set of images
    and will produce another set joining all elements of the 
    selected sets. It will validate that all sets are of the
    same type of elements (Micrographs, Particles or Volumes) 
    """
    _label = 'join sets'
    TYPE_CTF = 'CTFs'
    TYPE_VOLUME='Volumes'
    TYPE_VOLUME_INDEX = 3

    _unionTypes = ['Particles',
                   'Micrographs',
                   TYPE_CTF,
                   TYPE_VOLUME,
                   'Averages',
                   'All']

    def __init__(self, **kwargs):
        ProtSets.__init__(self, **kwargs)

        # We need to trace the changes of 'inputType' to
        # dynamically modify the property of pointerClass
        # of the 'inputSets' parameter
        def onChangeInputType():
            inputText = self.getEnumText('inputType')

            if inputText == 'All':
                pointerClass = 'EMSet'
            # elif inputText == 'CTFs + Micrographs':
            #     pointerClass = 'SetOfCTF'
            else:
                pointerClass = 'SetOf' + inputText
            # For relatively small set we usually want to include
            # the single element type, this will allow, for example
            # to union SetOfVolumes and Volumes in the final set
            if inputText in [self.TYPE_VOLUME]:
                pointerClass += ',%s' % inputText[:-1]  # remove last 's'
            elif inputText in [self.TYPE_CTF]:
                # remove last 's'
                pointerClass = '%s,CTFModel' % pointerClass[:-1]

            self.inputSetsParam.setPointerClass(pointerClass)

        self.inputType.trace(onChangeInputType)

    # -------------------------- DEFINE param functions ------------------------
    def _defineParams(self, form):
        form.addSection(label='Input')

        form.addParam('inputType', pwprot.params.EnumParam,
                      choices=self._unionTypes, default=5,  # All
                      label='Input type:',
                      help='Select the type of objects that you want to union.\n'
                           'Special case All will allow you to select any type.')
        self.inputSetsParam = form.addParam('inputSets', pwprot.params.MultiPointerParam,
                                            label="Input set", important=True,
                                            pointerClass='EMSet', minNumObjects=2, maxNumObjects=0,
                                            help='Select two or more sets (of micrographs, particles,'
                                                 ' volumes, etc.) to be united. If you select 3 sets '
                                                 'with 100, 200, 200 elements, the final set will '
                                                 'contain a total of 500 elements.')
        form.addParam('ignoreDuplicates', pwprot.params.BooleanParam,
                      default=False,
                      label='Remove duplicates?',
                      help='By default, duplicated items found (same ID) '
                           'within the input sets, will cause renumbering of all the '
                           'items ids in the output set. '
                           'This is the case for example when doing several '
                           'imports (which will cause ids overlapping) '
                           'but we really want to insert as new items in the '
                           'output. \n'
                           'On the other hand, items originated in a previous common '
                           'protocol (above in the workflow) might have identical items '
                           'and you would like to remove them. '
                           'Therefore, set this option to *Yes* to remove duplicates and keep only '
                           'one copy of the item (the first occurrence).')
        form.addParam('renumber', pwprot.params.BooleanParam, default=False,
                      expertLevel=pwprot.LEVEL_ADVANCED,
                      label="Force new ids",
                      help='Perform an automatic renumbering of ids to ensure all objects have unique ids. '
                           'This will mean new objects will not be associated to the old ones.')

        # TODO: See what kind of restrictions we add,
        # like "All sets should have the same sampling rate."

    # -------------------------- INSERT steps functions -----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep('createOutputStep')

    # --------------------------- STEPS functions ------------------------------
    def createOutputStep(self):

        set1 = self.inputSets[0].get()  # 1st set (we use it many times)

        # Read ClassName and create the corresponding EMSet (SetOfParticles...)
        try:
            if str(set1.getClassName()) is not Volume.__name__:
                outputSetFunction = getattr(self, "_create%s" % set1.getClassName())
            else:
                outputSetFunction = self._createSetOfVolumes
            outputSet = outputSetFunction()
        except Exception:
            outputSet = set1.createCopy(self._getPath())

        # Copy info from input sets (sampling rate, etc).
        if str(set1.getClassName()) is not Volume.__name__:
            outputSet.copyInfo(set1)  # all sets must have the same info as set1!
        else:
            outputSet.setSamplingRate(set1.getSamplingRate())

        # Renumber from the beginning if either the renumber option is selected
        # or we find duplicated ids in the sets
        cleanIds = not self.ignoreDuplicates.get() and self.duplicatedIds()

        # Warn in the log in case attributes will be lost
        allSetAttributes, commonAttrs = self.commonAttributes()
        warnings = self._getHeterogeneityWarning(allSetAttributes, commonAttrs)
        if warnings:
            self.info(warnings)

        # Always ingnore non common attributtes
        ignoreExtraAttributes = True

        # Get the 1st level attributes to be used for the copyAttributes
        copyAttrs = list()
        for attr in commonAttrs:
            if "." not in attr:
                copyAttrs.append(attr)

        self.info("Common attributes to all sets are: %s" % copyAttrs)

        idsList = {}
        setNum = 0
        for itemSet in self.inputSets:
            setNum += 1
            if str(itemSet.get().getClassName()) is not Volume.__name__:
                for obj in itemSet.get():
                    objId = obj.getObjId()
                    if self.ignoreDuplicates.get():
                        if objId in idsList:
                            continue
                        idsList[objId] =objId
                    # This is always TRUE, if stable we could remove the if and the else.
                    if ignoreExtraAttributes:
                        newObj = itemSet.get().ITEM_TYPE()
                        newObj.copyAttributes(obj, *copyAttrs)

                        self.cleanExtraAttributes(newObj, commonAttrs)
                        if not cleanIds or setNum == 1:
                            newObj.setObjId(objId)
                    else:
                        newObj = obj

                    if (cleanIds and setNum > 1) or self.renumber.get():
                        newObj.cleanObjId()

                    self._append(outputSet, newObj, sourceItem=obj)

            else:
                obj = itemSet.get()
                objId = obj.getObjId()
                if self.ignoreDuplicates.get():
                    if objId in idsList:
                        continue
                    idsList[objId] = objId
                newObj = obj
                if (cleanIds and setNum > 1) or self.renumber.get():
                    newObj.cleanObjId()
                outputSet.append(newObj)

        self._defineOutputs(outputSet=outputSet)
        for itemSet in self.inputSets:
            self._defineSourceRelation(itemSet, outputSet)

    # Overwrite SetOfCoordinates creation
    def _createSetOfCoordinates(self, suffix=''):
        coordSet = self.inputSets[0].get()
        micSet = coordSet.getMicrographs()
        return ProtSets._createSetOfCoordinates(self, micSet, suffix)

    def cleanExtraAttributes(self, obj, verifyAttrs, prefix=""):

        for attr, value in obj.getAttributesToStore():

            prefixedAttribute = prefix + attr

            if prefixedAttribute not in verifyAttrs:
                value._objDoStore = False
                self.info("%s will be lost." % attr)

            else:
                self.cleanExtraAttributes(value, verifyAttrs,
                                          prefixedAttribute + ".")

    # def getObjDict(self, includeClass=False, includeBasic=False):
    #     return super(ProtUnionSet, self).getObjDict(
    #         includeClass=includeClass, includeBasic=includeBasic)

    def duplicatedIds(self):
        """ Check if there are duplicated ids to renumber from
        the beginning. """
        usedIds = set()  # to keep track of the object ids we have already seen

        for item_pointer in self.inputSets:
            if str(item_pointer.get().getClassName()) is not Volume.__name__:
                for objId in item_pointer.get().getIdSet():
                    if objId in usedIds:
                        return True
            else:
                objId = item_pointer.get().getObjId()
                if objId in usedIds:
                    return True
            usedIds.add(objId)
        return False

    def getAllSetsAttributes(self):
        allSetsAttributes = list()
        for itemSet in self.inputSets:
            if str(itemSet.get().getClassName()) is not Volume.__name__:
                item = itemSet.get().getFirstItem()
            else:
                item = itemSet.get()
            attrs = set(item.getObjDict().keys())
            allSetsAttributes.append(attrs)

        return allSetsAttributes

    def commonAttributes(self):
        """ Compute the set of common attributes to all items within
        each input set. """
        commonAttrs = None
        allSetsAttributes = self.getAllSetsAttributes()

        for attrSet in allSetsAttributes:
            if commonAttrs is None:  # first time
                commonAttrs = attrSet
            else:
                commonAttrs = commonAttrs & attrSet

        return allSetsAttributes, list(commonAttrs)

    # -------------------------- INFO functions -------------------------------
    def _validate(self):
        # Are all inputSets from the same class?
        classes = {x.get().getClassName() for x in self.inputSets}
        if len(classes) > 1:
            return ["All objects should have the same type.",
                    "Types of objects found: %s" % ", ".join(classes)]
        if issubclass(type(self.inputSets[0].get()), emobj.SetOfClasses):
            return ["Is not possible to join different sets of classes.\n"
                    "If you want to join different representative, extract them "
                    "with the viewer and them run this protocol with the "
                    "resulting averages."]

        # Validate attributes like sampling rate or dimensions
        return self._checkSetsCompatibility()

    def _checkSetsCompatibility(self):
        """ Check if all input sets have a minimum compatible attributes """
        # Attributes to check
        attrs = {'sampling rates': 'getSamplingRate',
                 'dimensions': 'getDimensions'}
        errors = []
        # For each attribute
        for key, attr in attrs.items():

            # Intentional: we need a default value not None, since some
            # attributes could return None as a valid value.
            refValue = '?'

            # For pointer to a set
            for setPointer in self.inputSets:

                # Get the set:
                inputSet = setPointer.get()

                # If the set has the attribute
                if not hasattr(inputSet, attr):
                    break

                # Get the attribute and "call it" --> final ().
                setValue = getattr(inputSet, attr)()

                if refValue == '?':
                    refValue = setValue
                else:
                    if refValue != setValue:
                        errors.append("There are different %s among the input"
                                      " sets: %s and %s" % (key, refValue, setValue))
                        break

        return errors

    def _warnings(self):
        """ Warn about loosing info. """

        # Get all attributes "map"
        allSetsAttributes, commonAttributes = self.commonAttributes()

        return self._getHeterogeneityWarning(allSetsAttributes, commonAttributes)

    def _getHeterogeneityWarning(self, allSetsAttributes, commonAttributes):

        warnings = []
        # Use a set
        commonAttributes = set(commonAttributes)

        # Go through all sets attributes
        for index, setAttributes in enumerate(allSetsAttributes):
            setAttributes = set(setAttributes)
            # Get the difference
            lostAttributes = setAttributes - commonAttributes

            if len(lostAttributes) != 0:
                warnings.append("Set #%d will loose following "
                                "attributes:" % index)
                for attr in lostAttributes:
                    warnings.append(attr)

        if len(warnings):
            warnings.append("Your input sets have different attributes. "
                            "We will keep only the common ones. This may "
                            "cause the lost of important data like CFT, "
                            "alignment information,...")

        return  warnings

    def _summary(self):
        if not hasattr(self, 'outputSet'):
            return ["Protocol has not finished yet."]
        else:
            return ["We have merged the following sets:",
                    ", ".join(x.get().getNameId() for x in self.inputSets)]

    def _methods(self):
        return self._summary()


class ProtSplitSet(ProtSets):
    """ Protocol to split a set in two or more subsets.
    """
    _label = 'split sets'

    # -------------------------- DEFINE param functions -----------------------
    def _defineParams(self, form):
        form.addSection(label='Input')

        form.addParam('inputSet', pwprot.params.PointerParam,
                      pointerClass='EMSet',
                      label="Input set", important=True,
                      help='Select the set of elements (images, etc) that you '
                           'want to split.')

        form.addParam('numberOfSets', pwprot.params.IntParam, default=2,
                      label="Number of subsets",
                      help='Select how many subsets do you want to create.')

        form.addParam('randomize', pwprot.params.BooleanParam, default=False,
                      label="Randomize elements",
                      help='Put the elements at random in the different '
                           'subsets.')

    # Overwrite SetOfCoordinates creation
    def _createSetOfCoordinates(self, suffix=''):
        coordSet = self.inputSet.get()
        micSet = coordSet.getMicrographs()
        return ProtSets._createSetOfCoordinates(self, micSet, suffix)

    # -------------------------- INSERT steps functions -----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep('createOutputStep')

    # -------------------------- STEPS functions ------------------------------
    def createOutputStep(self):
        inputSet = self.inputSet.get()
        inputClassName = str(inputSet.getClassName())
        n = self.numberOfSets.get()
        # Create as many subsets as requested by the user
        try:
            outputSetFunction = getattr(self, "_create%s" % inputClassName)
            subsets = [outputSetFunction(suffix=str(i)) for i in range(1, n + 1)]
        except Exception:
            subsets = [inputSet.createCopy(self._getPath(), suffix=str(i))
                       for i in range(1, n + 1)]

        # Iterate over the elements in the input set and assign
        # to different subsets.
        elements = self.inputSet.get()

        ns = [len(elements) // n + (1 if i < len(elements) % n else 0)
              for i in range(n)]  # number of elements in each subset
        pos, i = 0, 0  # index of current subset and index of position inside it
        orderBy = 'RANDOM()' if self.randomize else 'id'

        for elem in elements.iterItems(orderBy=orderBy, direction='ASC'):
            if i >= ns[pos]:
                pos += 1
                i = 0
            self._append(subsets[pos], elem)
            i += 1

        key = 'output' + inputClassName.replace('SetOf', '') + '%02d'
        for i in range(1, n + 1):
            subset = subsets[i - 1]
            subset.copyInfo(inputSet)
            self._defineOutputs(**{key % i: subset})
            self._defineTransformRelation(inputSet, subset)

    # -------------------------- INFO functions -------------------------------
    def _validate(self):
        errors = []
        if self.inputSet.get().getSize() < self.numberOfSets:
            errors.append("The number of subsets requested is greater than")
            errors.append("the number of elements in the input set.")
        return errors

    def _summary(self):
        if not any(x.startswith('output') for x in dir(self)):
            return ["Protocol has not finished yet."]
        else:
            return ["We have split the set %s in %d sets." %
                    (self.inputSet.getName(), self.numberOfSets.get())]


class ProtSubSet(ProtSets):
    """    
    Create a set with the elements of an original set that are also
    referenced in another set.
    
    Usually there is a bigger set with all the elements, and a smaller
    one obtained from classification, cleaning, etc. The desired result
    is a set with the elements from the original set that are also present
    somehow in the smaller set (in the smaller set they may be downsampled
    or processed in some other way).
    
    Both sets should be of the same kind (micrographs, particles, volumes)
    or related (micrographs and CTFs for example).
    """
    _label = 'subset'
    SET_INTERSECTION = 0
    SET_DIFFERENCE = 1

    # -------------------------- DEFINE param functions -----------------------
    def _defineParams(self, form):
        form.addSection(label='Input')

        add = form.addParam  # short notation
        add('inputFullSet', pwprot.params.PointerParam, pointerClass='EMSet',
            label="Full set of items", important=True,
            help='Even if the operation can be applied to two arbitrary sets,\n'
                 'the most common use-case is to retrieve a subset of\n'
                 'elements from an original full set.\n'
                 '*Note*: the elements of the resulting set will be the same\n'
                 'ones as this input set.')
        add('chooseAtRandom', pwprot.params.BooleanParam, default=False,
            label="Make random subset",
            help='Choose elements randomly form the full set.')
        add('nElements', pwprot.params.IntParam, default=2,
            condition='chooseAtRandom',
            label="Number of elements",
            help='How many elements will be taken from the full set.')
        add('selectIds', pwprot.params.BooleanParam, default=False,
            condition='not chooseAtRandom',
            label="Make a subset from specific IDs",
            help="Choose specific elements form the full set.")
        add('range', pwprot.params.NumericRangeParam,
            label="IDs range or list",
            condition='selectIds and not chooseAtRandom',
            allowsNull=True,
            help='Select the IDs that will be the subset.\n'
                 'You have several ways to specify the IDs.\n'
                 'Example: \n'
                 '"1,3,5-8,17-20" -> [1,3, 5, 6, 7, 8, 17, 18, 19, 20]\n')
        add('inputSubSet', pwprot.params.PointerParam,
            pointerClass='EMSet', condition='not (chooseAtRandom or selectIds)',
            label="Other set",
            allowsNull=True,
            help='The elements present in this set will be used to pick \n'
                 'elements from the input full set.     \n'
                 'This means that the output set will contain elements with \n'
                 'exact the same information of input full set.\n\n'
                 'Set operation: if _intersection_ is used,\n'
                 'elements that are both in input and other set\n'
                 'will be included. If _difference_, elements that\n'
                 'are in input but not in other will picked.')
        add('setOperation', pwprot.params.EnumParam,
            condition='not (chooseAtRandom or selectIds)',
            default=self.SET_INTERSECTION,
            choices=['intersection', 'difference'],
            display=pwprot.params.EnumParam.DISPLAY_HLIST,
            label='Set operation',
            help='Set operation: if _intersection_ is used,\n'
                 'elements that are both in input and other set\n'
                 'will be included. If _difference_, elements that\n'
                 'are in input but not in other will picked.')

    # -------------------------- INSERT steps functions -----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep('createOutputStep')

    # -------------------------- STEPS functions ------------------------------
    def createOutputStep(self):
        inputFullSet = self.inputFullSet.get()

        inputClassName = inputFullSet.getClassName()

        try:
            outputSetFunction = getattr(self, "_create%s" % inputClassName)
            outputSet = outputSetFunction()
        except Exception:
            outputSet = inputFullSet.createCopy(self._getPath())

        outputSet.copyInfo(inputFullSet)

        if self.chooseAtRandom or self.selectIds:
            if self.chooseAtRandom:
                # Get all ids form iput set
                self.info("Creating subset from random positions from input set.")
                ids = set(random.sample(inputFullSet.getIdSet(), self.nElements.get()))
            else:
                self.info("Creating subset by range: %s" % self.range)
                ids = set(getListFromRangeString(self.range.get()))
        else:
            # Get the ids from both sets
            fullSetIds = inputFullSet.getIdSet()
            smallSetIds = self.inputSubSet.get().getIdSet()

            # The function to include an element or not
            # depends on the set operation
            # if it is 'intersection' we want that item is not None (found)
            # if it is 'difference' we want that item is None
            # (not found, different)
            if self.setOperation == self.SET_INTERSECTION:
                ids = fullSetIds.intersection(smallSetIds)
            else:
                ids = fullSetIds.difference(smallSetIds)

        progress = None
        nElements = len(ids)

        if nElements > 100000:  # show progressBar for large sets
            progress = ProgressBar(total=nElements, fmt=ProgressBar.NOBAR)
            progress.start()
            sys.stdout.flush()
            step = max(25000, nElements // 25000)

        i = 0

        for elem in inputFullSet.iterItems():
            if elem.getObjId() in ids:
                i += 1
                if progress and i % step == 0:
                    progress.update(i)
                self._append(outputSet, elem)

        if progress:
            progress.finish(printNewLine=True)

        if outputSet.getSize():
            key = 'output' + inputClassName.replace('SetOf', '')
            self._defineOutputs(**{key: outputSet})
            self._defineTransformRelation(inputFullSet, outputSet)
            if not (self.chooseAtRandom.get() or self.selectIds.get()):
                self._defineSourceRelation(self.inputSubSet, outputSet)
        else:
            self.summaryVar.set('Output was not generated. Resulting set '
                                'was EMPTY!!!')

    # Overwrite SetOfCoordinates creation
    def _createSetOfCoordinates(self, suffix=''):
        coordSet = self.inputFullSet.get()
        micSet = coordSet.getMicrographs()
        return ProtSets._createSetOfCoordinates(self, micSet, suffix)

    # -------------------------- INFO functions -------------------------------
    def _validate(self):
        """Make sure the input data make sense."""

        # Do not allow failing sets:
        notImplentedClasses = ['SetOfClasses2D', 'SetOfClasses3D',
                               'CoordinatesTiltPair']

        errors =[]
        if not self.chooseAtRandom and not self.selectIds and not self.inputSubSet.get():
            errors.append("Subsetting without ids or random selection needs the 'Other set' parameter.")

        if not self.inputFullSet.get():
            # Since is mandatory it will not validate
            # Stop validating since following validations need this set
            return errors

        c1 = self.inputFullSet.get().getClassName()
        if c1 in notImplentedClasses:
            errors.append("%s subset is not implemented." % c1)

        # First dispatch the easy case, where we choose elements at random.
        if self.chooseAtRandom:
            if self.nElements > self.inputFullSet.get().getSize():
                errors.append("Number of elements to choose cannot be bigger than",
                        "the number of elements in the set.")


        # Now the harder case: two sets. Check for compatible classes.

        # self.inputFullSet and self.inputSubSet .get().getClassName()
        # can be SetOf...:
        #   Alignment
        #   Angles
        #   Averages
        #   Classes
        #   ClassesVol
        #   Coordinates
        #   CTF
        #   Micrographs
        #   MovieParticles
        #   Movies
        #   Particles
        #   Volumes

        if not self.inputSubSet.get():
            # Stop validating since following validations need this set
            return errors
        
        c2 = self.inputSubSet.get().getClassName()
        if c2 in notImplentedClasses:
            errors.append("%s subset is not implemented." % c2)

        if c1 == c2:
            return errors

        # Avoid combinations that make no sense.
        for classA, classesIncompatible in [
            ('SetOfParticles',
             {'SetOfMicrographs', 'SetOfMovies', 'SetOfVolumes'}),
            ('SetOfCoordinates',
             {'SetOfMicrographs', 'SetOfMovies', 'SetOfVolumes'}),
            ('SetOfVolumes',
             {'SetOfMicrographs', 'SetOfMovies', 'SetOfParticles', 'SetOfCoordinates'})]:
            if ((c1 == classA and c2 in classesIncompatible) or
                    (c2 == classA and c1 in classesIncompatible)):
                errors.append("The full set and the subset are of incompatible classes",
                        "%s and %s." % (c1, c2))
        return errors

    def _summary(self):
        if self.summaryVar.hasValue():
            return [self.summaryVar.get()]

        key = 'output' + self.inputFullSet.get().getClassName().replace('SetOf', '')

        if not hasattr(self, key):
            return ["Protocol has not finished yet."]
        else:
            if self.setOperation == self.SET_INTERSECTION:
                return ["The elements of %s that also are referenced in %s" %
                        (self.inputFullSet.getName(), self.inputSubSet.getName()),
                        "are now in %s" % getattr(self, key).getName()]
            else:
                return ["%s has elements only present in %s." %
                        (getattr(self, key).getName(),
                         self.inputFullSet.getName())
                        ]


class ProtSubSetByMic(ProtSets):
    """
    Create a subset of those particles that come from a particular set of micrographs
    """
    _label = 'particles subset by micrograph'

    # --------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        form.addSection(label='Input')

        add = form.addParam  # short notation
        add('inputParticles', pwprot.params.PointerParam,
            pointerClass='SetOfParticles', label="Input particles",
            help='Set of particles from which the subset will be taken')
        add('inputMicrographs', pwprot.params.PointerParam,
            pointerClass='SetOfMicrographs', label="Input micrographs",
            help='Only the particles in this set of micrographs will be output')

    # --------------------------- INSERT steps functions ----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep('createOutputStep',
                                 self.inputParticles.getObjId(),
                                 self.inputMicrographs.getObjId())

    # --------------------------- STEPS functions -----------------------------
    def createOutputStep(self, partsId, micsId):
        inputParticles = self.inputParticles.get()
        inputMicrographs = self.inputMicrographs.get()

        outputSet = self._createSetOfParticles()
        outputSet.copyInfo(inputParticles)

        micIds = []

        for mic in inputMicrographs:
            micIds.append(mic.getObjId())

        for particle in inputParticles:
            if particle.getMicId() in micIds:
                outputSet.append(particle)

        self._defineOutputs(outputParticles=outputSet)
        self._defineTransformRelation(inputParticles, outputSet)

    # --------------------------- INFO functions ------------------------------
    def _validate(self):
        """Make sure the input data make sense, i.e. hasMicId.
        Thus they come from some Mic"""
        if not self.inputParticles.get().getFirstItem().hasMicId():
            return ['The _Input Particles_ must come from some Micrographs '
                    'of the workflow, i.e. particles must have micId.']

    def _summary(self):
        if not hasattr(self, 'outputParticles'):
            summary = ["Protocol has not finished yet."]
        else:
            summary = ['A subset of *%d* particles is made from a total of *%d*'
                       ' particles.' % (self.outputParticles.getSize(),
                                        self.inputParticles.get().getSize())]
        return summary


class ProtSubSetByCoord(ProtSets):
    """
    Create a subset of those particles that have a particular set of coordinates
    """
    _label = 'particles subset by coordinates'

    # --------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        form.addSection(label='Input')

        add = form.addParam  # short notation
        add('inputParticles', pwprot.params.PointerParam,
            pointerClass='SetOfParticles', label="Input particles",
            help='Set of particles from which the subset will be taken')
        add('inputCoordinates', pwprot.params.PointerParam,
            pointerClass='SetOfCoordinates', label="Input coordinates",
            help='Only the particles with this set of coordinates will be output')
        add('coordTolerance', pwprot.params.FloatParam,
            label='Coordinate tolerance (px)', default=0,
            help='Two coordinates are supposed to be the same if their X and Y distance'
                 ' is smaller or equal this value')

    # --------------------------- INSERT steps functions ----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep('createOutputStep',
                                 self.inputParticles.getObjId(),
                                 self.inputCoordinates.getObjId(),
                                 self.coordTolerance.get())

    # --------------------------- STEPS functions -----------------------------
    def createOutputStep(self, partsId, micsId, tolerance):
        inputParticles = self.inputParticles.get()
        inputCoordinates = self.inputCoordinates.get()

        outputSet = self._createSetOfParticles()
        outputSet.copyInfo(inputParticles)

        micCoordinates = {}
        for coord in inputCoordinates.iterCoordinates():
            micId = coord.getMicId()
            x, y = coord.getPosition()
            if micId not in micCoordinates:
                micCoordinates[micId] = []
            micCoordinates[micId].append((x, y))

        for particle in inputParticles:
            if particle.getMicId() in micCoordinates:
                x0, y0 = particle.getCoordinate().getPosition()
                okToAdd = False
                for x, y in micCoordinates[particle.getMicId()]:
                    if abs(x - x0) <= tolerance and abs(y - y0) <= tolerance:
                        okToAdd = True
                        break
                if okToAdd:
                    outputSet.append(particle)

        self._defineOutputs(outputParticles=outputSet)
        self._defineTransformRelation(inputParticles, outputSet)

    # --------------------------- INFO functions ------------------------------
    def _validate(self):
        """Make sure the input data make sense, i.e. hasMicId.
        Thus they come from some Mic"""
        if not self.inputParticles.get().getFirstItem().hasCoordinate():
            return ['The _Input Particles_ must have coordinates']

    def _summary(self):
        if not hasattr(self, 'outputParticles'):
            summary = ["Protocol has not finished yet."]
        else:
            summary = ['A subset of *%d* particles is made from a total of *%d*'
                       ' particles.' % (self.outputParticles.getSize(),
                                        self.inputParticles.get().getSize())]
        return summary




import unittest
from unittest.mock import patch
from pwem.protocols import ProtExtractCoords
from pwem.tests.utils import getSoPartMock, getSoMMock, getMicNameFromId


class TestExtractCoordinates(unittest.TestCase):
    """ Tests extract coordinates protocol mocking some behaviour"""

    def test_extractCoordinatesById(self):
        """ Tests ProtExtractCoords.extractCoordinates method using mic id for matching"""

        output = self._extractCoordinatesMocker(sop=getSoPartMock(name="inParts"),
                                                som=getSoMMock(name="inMics"))

        self.assertEqual(3, output.getSize(), "Wrong coordinates extraction")

        # Test missing mics
        output = self._extractCoordinatesMocker(sop=getSoPartMock(name="inPartsMissing"),
                                                som=getSoMMock(start=2, end=4, name="inMicsMissing"))

        self.assertEqual(2, output.getSize(), "Wrong coordinates extraction when missing mic ids")

    def test_extractCoordinatesByMicName(self):
        """ Tests ProtExtractCoords.extractCoordinates method using mic name for matching"""

        # Mic id wil range from 5-7
        som = getSoMMock(5, 7, "noIdButName")

        # Set micnames from 1-3
        for key, mic in som.items():
            mic.setMicName(getMicNameFromId(key - 4))

        output = self._extractCoordinatesMocker(sop=getSoPartMock(name="inPartsNoIdButNames"),
                                                som=som)

        self.assertEqual(3, output.getSize(), "Wrong coordinates extraction, matching by micname fails")

    @staticmethod
    def _extractCoordinatesMocker(sop, som):
        extractionProt = ProtExtractCoords(workingDir="/tmp")

        # Patch getInputParticles
        with patch.object(ProtExtractCoords, 'getInputParticles',
                          return_value=sop) as mock_method:
            # Patch getInputMics
            with patch.object(ProtExtractCoords, 'getInputMicrographs',
                              return_value=som) as mock_method2:
                # Not in streaming
                extractionProt.streamingModeOn = False

                return extractionProt.extractCoordinates()


if __name__ == '__main__':
    unittest.main()



# ***************************************************************************
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# ***************************************************************************/
import os

import pyworkflow.tests as pwtests

import pwem.protocols as emprot
import pwem.constants as emcts
import pwem.convert as emconv
from pwem.convert import Ccp4Header
from pyworkflow.object import Pointer


class TestImportBase(pwtests.BaseTest):
    @classmethod
    def setUpClass(cls):
        pwtests.setupTestProject(cls)
        cls.dsXmipp = pwtests.DataSet.getDataSet('xmipp_tutorial')
        cls.dsRelion = pwtests.DataSet.getDataSet('relion_tutorial')
        cls.dsEmx = pwtests.DataSet.getDataSet('emx')
        cls.dsMda = pwtests.DataSet.getDataSet('mda')
        cls.dsModBuild = pwtests.DataSet.getDataSet('model_building_tutorial')


class TestImportVolumes(TestImportBase):

    def test_import_volume(self):
        """ 1) Import single volume and set origin in default position (the
            volume will be centered in the center of coordinates)
        """
        args = {'filesPath': self.dsXmipp.getFile('volumes/'),
                'filesPattern': 'volume_1_iter_002.mrc',
                'setHalfMaps': False,
                'setOrigCoord': False,
                'samplingRate': 2.1,
                }

        # Id's should be set increasing from 1 if ### is not in the
        # pattern
        prot1 = self.newProtocol(emprot.ProtImportVolumes, **args)
        prot1.setObjLabel('import vol,\n default origin,\n chimera show '
                          'axis,\n vol origin in coord origin')
        self.launchProtocol(prot1)
        volume = prot1.outputVolume
        t = volume.getOrigin()
        x, y, z = t.getShifts()
        # x, y, z in Angstroms
        # Chimera will show (x, y, z) divided by the samplingRate
        # in pixels = (32, 32, 32)
        self.assertEqual(-67.2, x)
        self.assertEqual(-67.2, y)
        self.assertEqual(-67.2, z)

        # """ 2) Import single volume and set origin in userDefined position
        # """
        args = {'filesPath': self.dsXmipp.getFile('volumes/'),
                'filesPattern': 'volume_1_iter_002.mrc',
                'setHalfMaps': False,
                'setOrigCoord': True,
                'samplingRate': 2.1,
                'x': 16.8,
                'y': 33.6,
                'z': 50.4
                }

        # Id's should be set increasing from 1 if ### is not in the
        # pattern
        prot2 = self.newProtocol(emprot.ProtImportVolumes, **args)
        prot2.setObjLabel('import vol,\n user origin,\n chimera show axis')
        self.launchProtocol(prot2)
        volume = prot2.outputVolume
        t = volume.getOrigin()
        x, y, z = t.getShifts()

        # Assert path is relative
        self.assertFalse(os.path.isabs(volume.getFileName()),
                         "single volume path is not relative")

        # x, y, z in Angstroms
        # in Chimera we will see (x, y, z) divided by the samplingRate
        # in pixels = (8, 16, 24)
        self.assertEqual(-16.8, x)
        self.assertEqual(-33.6, y)
        self.assertEqual(-50.4, z)

        # """ 3) Import single volume and set origin in userDefined position (the
        #     volume will be centered in the center of coordinates because
        #     coordinates are the default ones)
        # """
        args = {'filesPath': self.dsXmipp.getFile('volumes/'),
                'filesPattern': 'volume_1_iter_002.mrc',
                'setHalfMaps': False,
                'setOrigCoord': True,
                'samplingRate': 2.1,
                'x': 67.2,
                'y': 67.2,
                'z': 67.2
                }

        # Id's should be set increasing from 1 if ### is not in the
        # pattern
        prot2 = self.newProtocol(emprot.ProtImportVolumes, **args)
        prot2.setObjLabel('import vol,\n user origin,\n chimera show axis')
        self.launchProtocol(prot2)
        volume = prot2.outputVolume
        t = volume.getOrigin()
        x, y, z = t.getShifts()

        # x, y, z in Angstroms
        # in Chimera we will see (x, y, z) divided by the samplingRate
        # in pixels = (8, 16, 24)
        self.assertEqual(-67.2, x)
        self.assertEqual(-67.2, y)
        self.assertEqual(-67.2, z)

        # """ 4) Import two volumes and set origin in default position
        # """
        args = {'filesPath': self.dsXmipp.getFile('volumes/'),
                'filesPattern': 'volume_*mrc',
                'setHalfMaps': False,
                'setOrigCoord': False,
                'samplingRate': 2.1,
                }

        # Id's should be set increasing from 1 if ### is not in the
        # pattern
        prot3 = self.newProtocol(emprot.ProtImportVolumes, **args)
        prot3.setObjLabel('import 2 vols,\n default origin,\n chimera no '
                          'axis')
        self.launchProtocol(prot3)
        for volume in prot3.outputVolumes:
            t = volume.getOrigin()
            x, y, z = t.getShifts()
            self.assertEqual(-67.2, x)
            self.assertEqual(-67.2, y)
            self.assertEqual(-67.2, z)
            # Assert path is relative
            self.assertFalse(os.path.isabs(volume.getFileName()),
                             "volume path is not relative")

        # """ 5) Import two volumes and set origin in userDefined position
        # """
        args = {'filesPath': self.dsXmipp.getFile('volumes/'),
                'filesPattern': 'volume_*mrc',
                'setHalfMaps': False,
                'setOrigCoord': True,
                'samplingRate': 2.1,
                'x': 16.8,
                'y': 33.6,
                'z': 50.4
                }

        # Id's should be set increasing from 1 if ### is not in the
        # pattern
        prot4 = self.newProtocol(emprot.ProtImportVolumes, **args)
        prot4.setObjLabel('import 2 vols,\n user origin,\n chimera no axis')
        self.launchProtocol(prot4)
        for volume in prot4.outputVolumes:
            t = volume.getOrigin()
            x, y, z = t.getShifts()
            self.assertEqual(-16.8, x)
            self.assertEqual(-33.6, y)
            self.assertEqual(-50.4, z)

        # """ 6) Import three volumes (mrc stack) and set origin in userDefined
        # position
        # """
        # Id's should be taken from filename
        args['filesPath'] = self.dsRelion.getFile('import/case2/'
                                                  'relion_volumes.mrc')
        args['filesPattern'] = ''
        prot2 = self.newProtocol(emprot.ProtImportVolumes, **args)
        prot2.setObjLabel('import 3 vols from mrc stack,\n user origin,'
                          '\n chimera no axis')
        self.launchProtocol(prot2)
        # Check the number of output volumes and dimensions
        self.assertEqual(3, prot2.outputVolumes.getSize())
        self.assertEqual(60, prot2.outputVolumes.getDim()[0])

        # """ 7) Import three volumes (spider stack) and set origin in
        # userDefined position
        # """
        # Id's should be taken from filename
        args['filesPath'] = self.dsRelion.getFile('import/case2/'
                                                  'relion_volumes.stk')
        args['filesPattern'] = ''
        prot3 = self.newProtocol(emprot.ProtImportVolumes, **args)
        prot3.setObjLabel('import 3 vols from spider stack\n user origin,'
                          '\n chimera no axis')
        self.launchProtocol(prot3)
        # Check the number of output volumes and dimensions
        self.assertEqual(3, prot3.outputVolumes.getSize())
        self.assertEqual(60, prot3.outputVolumes.getDim()[0])

        # """ 8)To test old data where volumes have no origin at all"""
        args = {'filesPath': self.dsXmipp.getFile('volumes/'),
                'filesPattern': 'volume_1_iter_002.mrc',
                'setHalfMaps': False,
                # 'setOrigCoord': False,
                'samplingRate': 2.1,
                }

        prot4 = self.newProtocol(emprot.ProtImportVolumes, **args)
        prot4.setObjLabel('import vol,\n no origin at all, legacy data,'
                          '\n chimera show '
                          'axis,\n vol origin in coord origin')
        self.launchProtocol(prot4)
        volume = prot4.outputVolume
        volume.setOrigin(None)
        # The volume has no origin
        t = volume.getOrigin(force=True)
        x, y, z = t.getShifts()
        # x, y, z in Angstroms
        # Chimera will show (x, y, z) divided by the samplingRate
        # in pixels = (32, 32, 32)
        self.assertEqual(-67.2, x)
        self.assertEqual(-67.2, y)
        self.assertEqual(-67.2, z)

        # TODO: associate origen coordinates from header file
        # args = {'filesPath': self.dsModBuild.getFile('volumes/1ake_4-5A.mrc'),
        #         'samplingRate': 1.5,
        #         'setDefaultOrigin': False,
        #         'x': ,
        #         'y': ,
        #         'z':
        #         }
        #
        # protImportVol = self.newProtocol(emprot.ProtImportVolumes, **args)
        # protImportVol.setObjLabel('import vol 1ake_4-5A,\n header origin,\n chimera show axis')
        # self.launchProtocol(protImportVol)
        # volume = protImportVol.outputVolume
        #
        # # Id's should be set increasing from 1 if ### is not in the
        # # pattern
        #
        # t = volume.getOrigin()
        # x, y, z = t.getShifts()
        # # x, y, z in Angstroms
        # # in Chimera we will see (x, y, z) divided by the samplingRate
        # self.assertEqual(-11.99, x)
        # self.assertEqual(7.88, y)
        # self.assertEqual(-10.90, z)

        # """ 9) Import 3D map from EMDB """
        from pwem.protocols.protocol_import.volumes import ProtImportVolumes
        args = {'importFrom': ProtImportVolumes.IMPORT_FROM_EMDB,
                'emdbId': 10676
                }

        prot5 = self.newProtocol(emprot.ProtImportVolumes, **args)
        prot5.setObjLabel('import vol,\n import from EMDB')
        self.launchProtocol(prot5)
        volume = prot5.outputVolume
        # volume.setOrigin(None)
        # The volume has no origin
        t = volume.getOrigin(force=True)
        x, y, z = t.getShifts()
        self.assertTrue(os.path.exists(prot5._getExtraPath('emd_10676.map')))
        # TODO: I should check origin but all 3D map
        # I have tried to download have origin =0 :-(
        # self.assertEqual(-67.2, x)
        # self.assertEqual(-67.2, y)
        # self.assertEqual(-67.2, z)

        # Test join sets work with volumes

        unionProt = self.newProtocol(emprot.ProtUnionSet,
                                     inputType=emprot.ProtUnionSet.TYPE_VOLUME_INDEX)

        # Set the input volumes
        unionProt.inputSets.append (Pointer(prot2, extended="outputVolumes.1"))
        unionProt.inputSets.append(Pointer(prot3, extended="outputVolumes.1"))
        unionProt.setObjLabel("Single volumes union")
        self.launchProtocol(unionProt)
        self.assertSetSize(unionProt.outputSet, 2, msg="Union of 2 volumes does not work.")


    def test_import_volume2(self):
        """
        Test to import a full map (Icosahedron) and two maps (half1 and half2)
        to compute the FSC
        """
        volFeatName = '/tmp/Icosahedron_map.txt'
        volMapNamefull = '/tmp/Icosahedron_map_full.mrc'
        volMapNamehalf1 = '/tmp/Icosahedron_map_half1.mrc'
        volMapNamehalf2 = '/tmp/Icosahedron_map_half2.mrc'
        createFeatVolume(volFeatName, volMapNamefull, sym=emcts.SYM_I222r)
        createFeatVolume(volFeatName, volMapNamehalf1, sym=emcts.SYM_I222r)
        createFeatVolume(volFeatName, volMapNamehalf2, sym=emcts.SYM_I222r)
        _samplingRate = 1.0

        args = {'filesPath': volMapNamefull,
                'filesPattern': '',
                'setHalfMaps': True,
                'half1map': volMapNamehalf1,
                'half2map': volMapNamehalf2,
                'samplingRate': _samplingRate
                }
        prot5 = self.newProtocol(emprot.ProtImportVolumes, **args)
        prot5.setObjLabel('import phantom icosahedron,\n half1 and half2')
        self.launchProtocol(prot5)
        volume = prot5.outputVolume
        volume.setOrigin(None)
        # The volume has no origin
        t = volume.getOrigin(force=True)
        x, y, z = t.getShifts()
        # x, y, z in Angstroms
        # Chimera will show (x, y, z) divided by the samplingRate
        # in pixels = ()
        self.assertEqual(-90.0, x)
        self.assertEqual(-90.0, y)
        self.assertEqual(-90.0, z)

    def test_import_volume3(self):
        """
        Test to import a full map (Icosahedron) and two maps (half1 and half2)
        to compute the FSC
        """
        volFeatName = '/tmp/Icosahedron_map.txt'
        volMapNamefull = '/tmp/Icosahedron_map_full.mrc'
        volMapNamehalf1 = '/tmp/Icosahedron_map_half1.vol'
        volMapNamehalf2 = '/tmp/Icosahedron_map_half2.map'
        createFeatVolume(volFeatName, volMapNamefull, sym=emcts.SYM_I222r)
        createFeatVolume(volFeatName, volMapNamehalf1, sym=emcts.SYM_I2n3)
        createFeatVolume(volFeatName, volMapNamehalf2, sym=emcts.SYM_In25r)
        _samplingRate = 1.0
        ORIG_X = 10
        ORIG_Y = 20
        ORIG_Z = 30
        args = {'filesPath': volMapNamefull,
                'filesPattern': '',
                'setHalfMaps': True,
                'half1map': volMapNamehalf1,
                'half2map': volMapNamehalf2,
                'samplingRate': _samplingRate,
                'setOrigCoord': True,
                'x': ORIG_X,
                'y': ORIG_Y,
                'z': ORIG_Z,
                }
        prot5 = self.newProtocol(emprot.ProtImportVolumes, **args)
        prot5.setObjLabel('import phantom icosahedron,\n half1 and half2 set origin')
        self.launchProtocol(prot5)
        volume = prot5.outputVolume
        self.assertVolumeOrigin(volume, -ORIG_X, -ORIG_Y, -ORIG_Z)

        # DO NOT REMOVE this print because it triggers the loading of the get of halfmaps
        print(volume.getHalfMaps())
        newHalf1Fn = volume.getHalfMaps().split(",")[0]
        self.assertNotEqual(volMapNamehalf1, newHalf1Fn)

        halfHeader = Ccp4Header(newHalf1Fn, readHeader=True)
        self.assertEqual((-ORIG_X, -ORIG_Y, -ORIG_Z), halfHeader.getOrigin(), "Origin in half map not set.")

    def assertVolumeOrigin(self, volume, x, y, z):
        # The volume has no origin
        t = volume.getOrigin()
        currx, curry, currz = t.getShifts()
        # x, y, z in Angstroms
        # Chimera will show (x, y, z) divided by the samplingRate
        # in pixels = ()
        self.assertEqual(x, currx)
        self.assertEqual(y, curry)
        self.assertEqual(z, currz)


def __runXmippProgram(program, args):
    """ Internal shortcut function to launch a Xmipp program.
    If xmipp not available o fails return False, else Tru"""
    try:
        from pwem import Domain
        xmipp3 = Domain.importFromPlugin('xmipp3', doRaise=True)
        xmipp3.Plugin.runXmippProgram(program, args)
    except ImportError:
        return False
    return True


def createFeatVolume(volFeatName, volMapName, sym=emcts.SYM_I222r, factor=1.):
    f = open(volFeatName, "w")
    f.write("""# Phantom description file, (generated with phantom help)
# General Volume Parameters:
#      Xdim      Ydim      Zdim   Background_Density Scale
   180 180 180 0 1.0
# Feature Parameters:
#Type  +/=  Density X_Center Y_Center Z_Center
""")
    icosahedron = emconv.Icosahedron(orientation=emcts.SCIPION_SYM_NAME[sym][1:])
    x = 0.
    y = 0.
    z = 0.
    f.write("# large sphere at the center\n")
    f.write("sph  + 1. %.3f %.3f %.3f %.3f\n" % (x, y, z, 36 * factor))
    f.write("# 5-fold\n")

    for i, vertice in enumerate(icosahedron.getVertices()):
        vertice = 55.0 * vertice
        f.write("sph  + 3 %.3f %.3f %.3f %.3f \n" %
                (vertice[0], vertice[1], vertice[2], 8.25 * factor))

    # print 3fold points
    f.write("# 3-fold\n")

    for _3fold in icosahedron.get3foldAxis():
        x, y, z = _3fold
        f.write("sph  + 0.8 %.3f %.3f %.3f %.3f \n" % (55.0 * x, 55.0 * y, 55.0 * z, 6 * factor))

    # print 2fold points
    f.write("# 2-fold\n")
    for _2fold in icosahedron.get2foldAxis():
        x, y, z = _2fold
        f.write("sph  + 0.7 %.3f %.3f %.3f %.3f\n" %
                (55.0 * x, 55.0 * y, 55.0 * z, 3. * factor))
    f.close()
    #    map
    program = "xmipp_phantom_create"
    args = '-i {featFile} -o {mapFile}'.format(
        featFile=volFeatName, mapFile=volMapName)
    __runXmippProgram(program, args)



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es) [1]
# * Authors:     Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [2]
# *
# * [1] Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# * [2] MRC Laboratory of Molecular Biology (MRC-LMB)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from os.path import exists
from glob import glob
try:
    from itertools import izip
except ImportError:
    izip = zip

from pyworkflow.protocol.params import (IntParam, PointerParam, FloatParam,
                                        BooleanParam, PathParam, EnumParam,
                                        LEVEL_ADVANCED)
import pyworkflow.utils as pwutils

from pwem import Domain
import pwem.emlib.metadata as md
from pwem.objects import CoordinatesTiltPair
from pwem.protocols import ProtParticlePicking

from .base import ProtImportFiles


class ProtImportCoordinates(ProtImportFiles, ProtParticlePicking):
    """ Protocol to import a set of coordinates """
    _label = 'import coordinates'

    IMPORT_FROM_AUTO = 0
    IMPORT_FROM_XMIPP = 1
    IMPORT_FROM_RELION = 2
    IMPORT_FROM_EMAN = 3
    IMPORT_FROM_DOGPICKER = 4
    IMPORT_FROM_SCIPION = 5

    OUTPUT_NAME='outputCoordinates'

    def _getImportChoices(self):
        """ Return a list of possible choices
        from which the import can be done.
        (usually packages formats such as: xmipp3, eman2, relion...etc.
        """
        return ['auto', 'xmipp', 'relion', 'eman', 'dogpicker', 'scipion']

    def _getDefaultChoice(self):
        return self.IMPORT_FROM_AUTO

    def _getFilesCondition(self):
        """ Return an string representing the condition
        when to display the files path and pattern to grab
        files.
        """
        return True

    # ---------------------- DEFINE param functions ---------------------------
    def _defineParams(self, form):

        ProtImportFiles._defineParams(self, form)

    def _defineImportParams(self, form):
        form.addParam('inputMicrographs', PointerParam,
                      pointerClass='SetOfMicrographs',
                      label='Input micrographs',
                      help='Select the micrographs for which you want to '
                           'import coordinates.')

        form.addParam('boxSize', IntParam, label='Box size')
        form.addParam('scale', FloatParam,
                      label='Scale', default=1,
                      condition='importFrom!=%d' % self.IMPORT_FROM_SCIPION,
                      help='Factor to scale coordinates')
        form.addParam('invertX', BooleanParam, default=False,
                      condition='importFrom!=%d' % self.IMPORT_FROM_SCIPION,
                      label='Invert X')
        form.addParam('invertY', BooleanParam, default=False,
                      condition='importFrom!=%d' % self.IMPORT_FROM_SCIPION,
                      label='Invert Y',
                      help='Invert Y for EMAN coordinates taken on dm3 or'
                           ' tif micrographs')

    # ------------------- INSERT steps functions ------------------------------
    def _insertAllSteps(self):
        importFrom = self.importFrom.get()
        self._insertFunctionStep('createOutputStep', importFrom,
                                 self.filesPath.get())

    # ------------------ STEPS functions --------------------------------------
    def createOutputStep(self, importFrom, *args):

        # scipion imports have a predefined format (see dataimport)
        if importFrom == self.IMPORT_FROM_SCIPION:
            from .dataimport import ScipionImport
            self.importFilePath = self.filesPath.get('').strip()
            sImport = ScipionImport(self, self.importFilePath)
            # this function defines outputs and relations
            sImport.importCoordinates(self.inputMicrographs)
        else:
            # Pass the pointer with extended.
            coordsSet = self._createSetOfCoordinates(self.inputMicrographs)
            coordsSet.setBoxSize(self.boxSize.get())
            ci = self.getImportClass()
            for coordFile, fileId in self.iterFiles():
                mic = self.getMatchingMic(coordFile, fileId)
                if mic is not None:
                    def addCoordinate(coord):
                        coord.setMicrograph(mic)
                        self.correctCoordinatePosition(coord)
                        coordsSet.append(coord)
                    # Parse the coordinates in the given format for this micrograph
                    ci.importCoordinates(coordFile, addCoordinate=addCoordinate)

            self._defineOutputs(**{self.OUTPUT_NAME:coordsSet})
            self._defineSourceRelation(self.inputMicrographs, coordsSet)

    # ---------------- INFO functions -----------------------------------------
    def _summary(self):
        summary = []

        if not hasattr(self, self.OUTPUT_NAME):
            msg = 'Output coordinates not ready yet'
        else:
            msg = "%s  coordinates from " % self.outputCoordinates.getSize()
            msg += "micrographs %s " % self.getObjectTag('inputMicrographs')
            importFrom = self._getImportChoices()[self.getImportFrom()]
            msg += "were imported using %s format." % importFrom
            if self.scale.get() != 1.:
                msg += " Scale factor %0.2f was applied." % self.scale
            if self.invertX.get():
                msg += " X coordinate was inverted."
            if self.invertY.get():
                msg += " Y coordinate was inverted."

            summary.append(msg)
            summary.append("Output coordinates: %s." % self.getObjectTag(self.OUTPUT_NAME))
        return summary

    def _methods(self):
        return self._summary()

    # ------------------ UTILS functions --------------------------------------
    def getImportClass(self):
        """ Return the class in charge of importing the files. """
        filesPath = self.filesPath.get()
        importFrom = self.getImportFrom()

        if importFrom == self.IMPORT_FROM_XMIPP:
            XmippImport = Domain.importFromPlugin('xmipp3.convert', 'XmippImport',
                                                  'Xmipp is needed to import .xmd files',
                                                  doRaise=True)
            return XmippImport(self, filesPath)

        elif importFrom == self.IMPORT_FROM_RELION:
            RelionImport = Domain.importFromPlugin('relion.convert', 'RelionImport',
                                                   errorMsg='Relion is needed to import .star files',
                                                   doRaise=True)
            return RelionImport(self, filesPath)

        elif importFrom == self.IMPORT_FROM_EMAN:
            EmanImport = Domain.importFromPlugin('eman2.convert', 'EmanImport',
                                                 errorMsg='Eman is needed to import .json or '
                                                          '.box files',
                                                 doRaise=True)
            return EmanImport(self, None)

        elif importFrom == self.IMPORT_FROM_DOGPICKER:
            DogpickerImport = Domain.importFromPlugin('appion.convert', 'DogpickerImport',
                                                      errorMsg='appion plugin is needed to import '
                                                               'dogpicker files',
                                                      doRaise=True)
            return DogpickerImport(self)

        else:
            self.importFilePath = ''
            return None

    def getImportFrom(self):
        importFrom = self.importFrom.get()
        if importFrom == self.IMPORT_FROM_AUTO:
            importFrom = self.getFormat()
        return importFrom

    def getFormat(self):
        for coordFile, _ in self.iterFiles():
            if coordFile.endswith('.pos'):
                return self.IMPORT_FROM_XMIPP
            if coordFile.endswith('.star'):
                return self.IMPORT_FROM_RELION
            if coordFile.endswith('.json') or coordFile.endswith('.box'):
                return self.IMPORT_FROM_EMAN
        return -1

    def getInputMicrographs(self):
        return self.inputMicrographs.get()

    def getMatchingMic(self, coordFile, fileId):
        """ Given a coordinates file check if there is a micrograph
        that this files matches.
        """
        micSet = self.inputMicrographs.get()

        if fileId is None:
            coordBase = pwutils.removeBaseExt(coordFile)
            for mic in micSet:
                micBase = pwutils.removeBaseExt(mic.getFileName())
                # temporal use of in
                if coordBase in micBase or micBase in coordBase:
                    return mic
            return None
        else:
            return micSet[fileId]

    def correctCoordinatePosition(self, coord):
        mic = coord.getMicrograph()
        scaleFactor = self.scale.get()
        x = coord.getX()
        y = coord.getY()
        if scaleFactor != 1.:
            x = coord.getX() * scaleFactor
            y = coord.getY() * scaleFactor
        if self.invertX:
            width = mic.getDim()[0]
            x = width - x
        if self.invertY:
            height = mic.getDim()[1]
            y = height - y
        coord.setX(x)
        coord.setY(y)

    def getDefaultBoxSize(self):
        """ This function is used by the wizard to estimate the box size. """
        boxSize = None
        ci = self.getImportClass()
        if hasattr(ci, 'getBoxSize'):
            boxSize = ci.getBoxSize([f for f, _ in self.iterFiles()][0])
            if boxSize is not None and self.scale != 1.:
                boxSize = int(boxSize * self.scale.get())

        return boxSize


class ProtImportCoordinatesPairs(ProtImportFiles):
    """ Protocol to import a set of tilt pair coordinates """
    _label = 'import coordinate pairs'

    IMPORT_FROM_XMIPP = 0
    IMPORT_FROM_EMAN = 1

    def _getImportChoices(self):
        """ Return a list of possible choices
        from which the import can be done.
        (usually packages formats such as: xmipp3, eman2..etc.)
        """
        return ['xmipp', 'eman']

    def _getDefaultChoice(self):
        return self.IMPORT_FROM_XMIPP

    # --------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        importChoices = self._getImportChoices()

        form.addSection(label='Import')
        form.addParam('importFrom', EnumParam,
                      choices=importChoices, default=self._getDefaultChoice(),
                      label='Import from',
                      help='Select the type of import.\n'
                           '_Auto_ - detects coordinate file type by extension.\n'
                           '_Xmipp_ - provide *.pos files\n'
                           '_Eman_ - provide info/*.json files')
        form.addParam('copyFiles', BooleanParam, default=False,
                      expertLevel=LEVEL_ADVANCED,
                      label="Copy files?",
                      help="By default the files are not copied into the "
                           "project to avoid data duplication and to save "
                           "disk space. Instead of copying, symbolic links are "
                           "created pointing to original files. This approach "
                           "has the drawback that if the project is moved to "
                           "another computer, the links need to be restored.")
        form.addParam('xmippMdFn', PathParam, condition='importFrom==0',
                      important=True,
                      label='Micrograph .xmd file',
                      help='Provide input_micrographs.xmd file that contains '
                           'tilt angles information. This file is usually '
                           'created alongside the coordinates.')
        form.addParam('patternUntilted', PathParam, label=pwutils.Message.LABEL_PATTERNU,
                      help=pwutils.Message.TEXT_PATTERN)
        form.addParam('patternTilted', PathParam, label=pwutils.Message.LABEL_PATTERNT,
                      help=pwutils.Message.TEXT_PATTERN)
        form.addParam('inputMicrographsTiltedPair', PointerParam,
                      pointerClass='MicrographsTiltPair',
                      label='Input tilt pair micrographs',
                      help='Select the tilt pair micrographs for which you want '
                           'to import coordinates.')
        form.addParam('boxSize', IntParam, label='Box size')
        form.addParam('scale', FloatParam,
                      label='Scale', default=1,
                      help='Factor to scale coordinates')
        form.addParam('invertX', BooleanParam, default=False,
                      label='Invert X')
        form.addParam('invertY', BooleanParam, default=False,
                      label='Invert Y')

    # --------------------------- INSERT steps functions ----------------------
    def _insertAllSteps(self):
        self.micsFn = self._getExtraPath('input_micrographs.xmd')
        self._insertFunctionStep('createOutputStep')

    # --------------------------- STEPS functions -----------------------------
    def createOutputStep(self):
        micTiltPairs = self.getInputMicrographs()
        # Get the converted input micrographs in Xmipp format
        writeSetOfMicrographsPairs = Domain.importFromPlugin('xmipp3.convert',
                                                             'writeSetOfMicrographsPairs')
        writeSetOfMicrographsPairs(micTiltPairs.getUntilted(),
                                   micTiltPairs.getTilted(),
                                   self.micsFn)
        uCoordSet, tCoordSet, anglesSet = self.readCoordinates()
        # Create CoordinatesTiltPair object
        coordsSet = self._createCoordinatesTiltPair(micTiltPairs,
                                                    uCoordSet, tCoordSet,
                                                    anglesSet, suffix='')
        self._defineOutputs(outputCoordinatesTiltPair=coordsSet)
        self._defineSourceRelation(self.inputMicrographsTiltedPair, coordsSet)

    # --------------------------- INFO functions ------------------------------
    def _validate(self):
        errors = []
        if not self.patternUntilted.get() or not self.patternTilted.get():
            errors.append(pwutils.Message.ERROR_PATTERN_EMPTY)
        else:
            filePathsUntilted = glob(pwutils.expandPattern(self.patternUntilted.get()))
            filePathsTilted = glob(pwutils.expandPattern(self.patternTilted.get()))

            if len(filePathsUntilted) == 0 or len(filePathsTilted) == 0:
                errors.append(pwutils.Message.ERROR_PATTERN_FILES)

        return errors

    def _summary(self):
        summary = []

        if not hasattr(self, 'outputCoordinatesTiltPair'):
            msg = 'Output tilt pair coordinates not ready yet'
        else:
            msg = "%s  coordinate pairs from micrographs %s were imported using %s format." % (
                self.outputCoordinatesTiltPair.getSize(),
                self.getObjectTag('inputMicrographsTiltedPair'),
                self._getImportChoices()[self.importFrom.get()])
            if self.scale.get() != 1.:
                msg += " Scale factor %0.2f was applied." % self.scale
            if self.invertX.get():
                msg += " X coordinate was inverted."
            if self.invertY.get():
                msg += " Y coordinate was inverted."

            summary.append(msg)
            summary.append("Output coordinates: %s." % self.getObjectTag('outputCoordinatesTiltPair'))
        return summary

    def _methods(self):
        return self._summary()

    # --------------------------- UTILS functions -----------------------------
    def getImportClass(self):
        """ Return the class in charge of importing the files. """
        importFrom = self.importFrom.get()

        if importFrom == self.IMPORT_FROM_XMIPP:
            XmippImport = Domain.importFromPlugin('xmipp3.convert', 'XmippImport',
                                                  'Xmipp is needed to import .xmd files',
                                                  doRaise=True)
            return XmippImport(self, None)
        else:  # import from EMAN
            EmanImport = Domain.importFromPlugin('eman2.convert', 'EmanImport',
                                                 errorMsg='Eman is needed to import .json or '
                                                          '.box files',
                                                 doRaise=True)
            return EmanImport(self, None)

    def getInputMicrographs(self):
        return self.inputMicrographsTiltedPair.get()

    def getMatchingCoord(self, micU, micT):
        """ Given a pair of mics check if there are coordinates
        that match.
        """
        mics = [micU, micT]
        patterns = [self.patternUntilted.get(), self.patternTilted.get()]
        coords = []

        for mic, pattern in izip(mics, patterns):
            micFnBase = pwutils.removeBaseExt(mic.getFileName())
            for coordFn in self._iterFiles(pattern):
                if coordFn not in coords:
                    coorBase = pwutils.removeBaseExt(coordFn).replace('_info', '')
                    if coorBase in micFnBase or micFnBase in coorBase:
                        coords.append(coordFn)

        if len(coords) == 2:
            return coords[0], coords[1]
        else:
            return False, False

    def correctCoordinatePosition(self, coord):
        mic = coord.getMicrograph()
        scaleFactor = self.scale.get()
        x = coord.getX()
        y = coord.getY()
        if scaleFactor != 1.:
            x = coord.getX() * scaleFactor
            y = coord.getY() * scaleFactor
        if self.invertX:
            width = mic.getDim()[0]
            x = width - x
        if self.invertY:
            height = mic.getDim()[1]
            y = height - y
        coord.setX(x)
        coord.setY(y)

    def getDefaultBoxSize(self):
        """ This function is used by the wizard to estimate the box size. """
        ci = self.getImportClass()
        pattern = self.patternUntilted.get()
        if hasattr(ci, 'getBoxSize'):
            boxSize = ci.getBoxSize([f for f in self._iterFiles(pattern)][0])
            if boxSize is not None and self.scale != 1.:
                boxSize = int(boxSize * self.scale.get())
            return boxSize

        boxSize = 100
        scale = self.scale.get()
        boxSize = int(boxSize * scale)

        return boxSize

    def readCoordinates(self):
        micTiltPairs = self.getInputMicrographs()
        ci = self.getImportClass()
        uSet = micTiltPairs.getUntilted()
        tSet = micTiltPairs.getTilted()
        # Create Untilted and Tilted SetOfCoordinates
        uCoordSet = self._createSetOfCoordinates(uSet, suffix='Untilted')
        tCoordSet = self._createSetOfCoordinates(tSet, suffix='Tilted')
        anglesSet = self._createSetOfAngles()

        def _importCoords(uCoordSet, tCoordSet):
            for micU, micT in izip(uSet, tSet):
                coordFnU, coordFnT = self.getMatchingCoord(micU, micT)
                if coordFnU and coordFnT:

                    def addCoordinateU(coord):
                        coord.setMicrograph(micU)
                        self.correctCoordinatePosition(coord)
                        uCoordSet.append(coord)

                    ci.importCoordinates(coordFnU, addCoordinate=addCoordinateU)

                    def addCoordinateT(coord):
                        coord.setMicrograph(micT)
                        self.correctCoordinatePosition(coord)
                        tCoordSet.append(coord)

                    ci.importCoordinates(coordFnT, addCoordinate=addCoordinateT)

                    def addAngles(ang):
                        anglesSet.append(ang)

                    if self.importFrom.get() == self.IMPORT_FROM_EMAN:
                        ci.importAngles(coordFnT, addAngles=addAngles)

        _importCoords(uCoordSet, tCoordSet)
        boxSize = self.boxSize.get()
        uCoordSet.setBoxSize(boxSize)
        tCoordSet.setBoxSize(boxSize)

        uCoordSet.write()
        tCoordSet.write()

        if self.importFrom.get() == self.IMPORT_FROM_XMIPP:
            readAnglesFromMicrographs = Domain.importFromPlugin('xmipp3.convert',
                                                                'readAnglesFromMicrographs')
            if exists(self.xmippMdFn.get()):
                checkAngles = self._compareMicPairs(self.xmippMdFn, uSet, tSet)
                if checkAngles:
                    readAnglesFromMicrographs(self.xmippMdFn, anglesSet)
                else:
                    self.warning('Angles for some micrographs were not found, '
                                 'skipping angles import')
            else:
                self.warning('Micrograph xmd file not provided, so tilt angles will '
                             'not be imported')
        anglesSet.write()

        return uCoordSet, tCoordSet, anglesSet

    def _iterFiles(self, pattern):
        filePaths = glob(pwutils.expandPattern(pattern))
        for fn in filePaths:
            yield fn

    def _compareMicPairs(self, micFn, uSet, tSet):
        # compare micFn input file and input micsSet
        micMd = md.MetaData(micFn)
        micFnDict = {}
        inputMicsDict = {}
        XmippMdRow = Domain.importFromPlugin('xmipp3', 'XmippMdRow')
        for objId in micMd:
            row = XmippMdRow()
            row.readFromMd(micMd, objId)
            micUFn = pwutils.removeBaseExt(row.getValue(md.MDL_MICROGRAPH))
            micTFn = pwutils.removeBaseExt(row.getValue(md.MDL_MICROGRAPH_TILTED))
            micFnDict[micUFn] = micTFn

        for micU, micT in izip(uSet, tSet):
            inputMicsDict[pwutils.removeBaseExt(micU.getFileName())] = pwutils.removeBaseExt(micT.getFileName())

        for micKey in inputMicsDict:
            if micKey not in micFnDict:
                return False
        return True

    # all funcs below are required for tilt picker to work properly

    def __getOutputSuffix(self):
        maxCounter = -1
        for attrName, _ in self.iterOutputAttributes(CoordinatesTiltPair):
            suffix = attrName.replace('outputCoordinatesTiltPair', '')
            try:
                counter = int(suffix)
            except Exception as ex:
                counter = 1  # when there is not number assume 1
            maxCounter = max(counter, maxCounter)

        return str(maxCounter + 1) if maxCounter > 0 else ''  # empty if not outputs

    def getSummary(self, coordSet):
        summary = []
        summary.append("Number of particle pairs picked: %s" % coordSet.getSize())
        summary.append("Particle size: %s" % coordSet.getBoxSize())
        return "\n".join(summary)

    def _getBoxSize(self):
        """ Redefine this function to set a specific box size to the output
        coordinates untilted and tilted.
        """
        return None

    def _readAngles(self, micsFn, suffix=''):
        # Read Angles from input micrographs
        anglesSet = self._createSetOfAngles(suffix=suffix)
        readAnglesFromMicrographs = Domain.importFromPlugin('xmipp3.convert',
                                                            'readAnglesFromMicrographs')
        readAnglesFromMicrographs(micsFn, anglesSet)
        anglesSet.write()
        return anglesSet

    def _readCoordinates(self, coordsDir, suffix=''):
        micTiltPairs = self.inputMicrographsTiltedPair.get()
        uSuffix = 'Untilted' + suffix
        tSuffix = 'Tilted' + suffix
        uSet = micTiltPairs.getUntilted()
        tSet = micTiltPairs.getTilted()
        # Create Untilted and Tilted SetOfCoordinates
        readSetOfCoordinates = Domain.importFromPlugin('xmipp3.convert',
                                                       'readSetOfCoordinates')
        uCoordSet = self._createSetOfCoordinates(uSet, suffix=uSuffix)
        readSetOfCoordinates(coordsDir, uSet, uCoordSet)
        uCoordSet.write()
        tCoordSet = self._createSetOfCoordinates(tSet, suffix=tSuffix)
        readSetOfCoordinates(coordsDir, tSet, tCoordSet)
        tCoordSet.write()
        boxSize = self._getBoxSize()
        if boxSize:
            uCoordSet.setBoxSize(boxSize)
            tCoordSet.setBoxSize(boxSize)

        return uCoordSet, tCoordSet

    def registerCoords(self, coordsDir, store=True, readFromExtra=True):
        micTiltPairs = self.inputMicrographsTiltedPair.get()
        suffix = self.__getOutputSuffix()

        uCoordSet, tCoordSet = self._readCoordinates(coordsDir, suffix)

        if readFromExtra:
            micsFn = self._getExtraPath('input_micrographs.xmd')
        else:
            micsFn = self._getPath('input_micrographs.xmd')

        anglesSet = self._readAngles(micsFn, suffix)
        # Create CoordinatesTiltPair object
        outputset = self._createCoordinatesTiltPair(micTiltPairs,
                                                    uCoordSet, tCoordSet,
                                                    anglesSet, suffix)
        summary = self.getSummary(outputset)
        outputset.setObjComment(summary)
        outputName = 'outputCoordinatesTiltPair' + suffix
        outputs = {outputName: outputset}
        self._defineOutputs(**outputs)
        self._defineSourceRelation(self.inputMicrographsTiltedPair, outputset)
        if store:
            self._store()



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os
from os.path import join
from glob import glob
import re
from datetime import datetime

import pyworkflow.utils as pwutils
import pyworkflow.protocol.params as params

from pwem.protocols import EMProtocol


class ProtImport(EMProtocol):
    """ Base class for other all Import protocols. """


class ProtImportFiles(ProtImport):
    """ Base class for other Import protocols. 
    All imports protocols will have:
    1) Several options to import from (_getImportOptions function)
    2) First option will always be "from files". (for this option 
      files with a given pattern will be retrieved  and the ### will 
      be used to mark an ID part from the filename.
      - For each file a function to process it will be called
        (_importFile(fileName, fileId))
    """
    IMPORT_FROM_FILES = 0

    # --------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        importChoices = self._getImportChoices()
        filesCondition = self._getFilesCondition()

        form.addSection(label='Import')

        if len(importChoices) > 1:  # not only from files
            form.addParam('importFrom', params.EnumParam,
                          choices=importChoices, default=self._getDefaultChoice(),
                          label='Import from',
                          help='Select the type of import.')
        else:
            form.addHidden('importFrom', params.EnumParam,
                           choices=importChoices, default=self.IMPORT_FROM_FILES,
                           label='Import from',
                           help='Select the type of import.')
        form.addParam('filesPath', params.PathParam,
                      condition=filesCondition,
                      label="Files directory",
                      help="Directory with the files you want to import.\n\n"
                           "The path can also contain wildcards to select"
                           "from several folders. \n\n"
                           "Examples:\n"
                           "  ~/project/data/day??_files/\n"
                           "Each '?' represents one unknown character\n\n"
                           "  ~/project/data/day*_files/\n"
                           "'*' represents any number of unknown characters\n\n"
                           "  ~/project/data/day##_files/\n"
                           "'##' represents two digits that will be used as "
                           "file ID\n\n"
                           "NOTE: wildcard characters ('*', '?', '#') "
                           "cannot appear in the actual path.)")
        form.addParam('filesPattern', params.StringParam,
                      label='Pattern',
                      condition=filesCondition,
                      help="Pattern of the files to be imported.\n\n"
                           "The pattern can contain standard wildcards such as\n"
                           "*, ?, etc, or special ones like ### to mark some\n"
                           "digits in the filename as ID.\n\n"
                           "NOTE: wildcards and special characters "
                           "('*', '?', '#', ':', '%') cannot appear in the "
                           "actual path.")
        form.addParam('copyFiles', params.BooleanParam, default=False,
                      expertLevel=params.LEVEL_ADVANCED,
                      label="Copy files?",
                      help="By default the files are not copied into the "
                           "project to avoid data duplication and to save "
                           "disk space. Instead of copying, symbolic links are "
                           "created pointing to original files. This approach "
                           "has the drawback that if the project is moved to "
                           "another computer, the links need to be restored.")

        self._defineImportParams(form)

        self._defineAcquisitionParams(form)

        form.addSection('Streaming')

        form.addParam('dataStreaming', params.BooleanParam, default=False,
                      label="Process data in streaming?",
                      help="Select this option if you want import data as it is "
                           "generated and process on the fly by next protocols. "
                           "In this case the protocol will keep running to check "
                           "new files and will update the output Set, which can "
                           "be used right away by next steps.")

        form.addParam('timeout', params.IntParam, default=43200,
                      condition='dataStreaming',
                      label="Timeout (secs)",
                      help="Interval of time (in seconds) after which, if no new file "
                           "is detected, the protocol will end. When finished, "
                           "the output Set will be closed and no more data will be "
                           "added to it. \n"
                           "Note 1:  The default value is  high (12 hours) to avoid "
                           "the protocol finishes during the acquisition of the "
                           "microscope. You can also stop it from right click and press "
                           "STOP_STREAMING.\n"
                           "Note 2: If you're using individual frames when importing "
                           "movies, the timeout won't be refreshed until a whole "
                           "movie is stacked.")

        form.addParam('fileTimeout', params.IntParam, default=30,
                      condition='dataStreaming',
                      label="File timeout (secs)",
                      help="Interval of time (in seconds) after which, if a file has "
                           "not changed, we consider it as a new file. \n")

        self._defineBlacklistParams(form)

    def _defineImportParams(self, form):
        """ Override to add options related to the different types
        of import that are allowed by each protocol.
        """
        pass

    def _defineAcquisitionParams(self, form):
        """ Override to add options related to acquisition info.
        """
        pass

    def _defineBlacklistParams(self, form):
        """ Override to add options related to blacklist info.
        """
        pass

    def _getDefaultChoice(self):
        return self.IMPORT_FROM_FILES

    # --------------------------- INFO functions ------------------------------
    def _validate(self):
        errors = []
        if self.importFrom == self.IMPORT_FROM_FILES:
            if not self.getPattern():
                errors.append("The path and pattern can not be both empty!!!")
            else:
                # Just check the number of files matching the pattern
                self.getMatchFiles()
                if self.numberOfFiles == 0:
                    errors.append("There are no files matching the pattern %s"
                                  % self.getPattern())

        return errors

    # --------------------------- BASE methods to be overwritten ----------------
    def _getImportChoices(self):
        """ Return a list of possible choices
        from which the import can be done.
        (usually packages formats such as: xmipp3, eman2, relion...etc.
        """
        return ['files']

    def _getFilesCondition(self):
        """ Return an string representing the condition
        when to display the files path and pattern to grab
        files.
        """
        return '(importFrom == %d)' % self.IMPORT_FROM_FILES

    # --------------------------- UTILS functions -----------------------------
    def getPattern(self):
        """ Expand the pattern using environ vars or username
        and also replacing special character # by digit matching.
        """
        self._idRegex = None
        filesPath = self.filesPath.get('').strip()
        filesPattern = self.filesPattern.get('').strip()

        if filesPattern:
            fullPattern = join(filesPath, filesPattern)
        else:
            fullPattern = filesPath

        pattern = pwutils.expandPattern(fullPattern.replace("$", ""))
        match = re.match('[^#]*(#+)[^#]*', pattern)

        if match is not None:
            g = match.group(1)
            n = len(g)
            # prepare regex pattern - place ids, handle *, handle ?
            idregex = pattern.replace(g, '(%s)' % ('[0-9]'*n))
            idregex = idregex.replace('*', '.*')
            idregex = idregex.replace('?', '.')
            self._idRegex = re.compile(idregex)
            pattern = pattern.replace(g, '[0-9]'*n)

        return pattern

    def getMatchFiles(self, pattern=None):
        """ Return a sorted list with the paths of files that
        matched the pattern.
        """
        if pattern is None:
            pattern = self.getPattern()

        filePaths = glob(pattern)
        filePaths.sort()
        self.numberOfFiles = len(filePaths)

        return filePaths

    def getCopyOrLink(self):
        # Set a function to copyFile or createLink
        # depending in the user selected option 
        if self.copyFiles:
            return pwutils.copyFile
        else:
            return pwutils.createAbsLink

    def fileModified(self, fileName, fileTimeout):
        """ Check if the fileName modification time is less
        than a given timeout.
        Params:
            fileName: input filename that will be checked.
            fileTimeout: timeout """
        self.debug('Checking file: %s' % fileName)
        mTime = datetime.fromtimestamp(os.path.getmtime(fileName))
        delta = datetime.now() - mTime
        self.debug('   Modification time: %s' % pwutils.prettyTime(mTime))
        self.debug('   Delta: %s' % pwutils.prettyDelta(delta))

        return delta < fileTimeout

    def isBlacklisted(self, fileName):
        """ Overwrite in subclasses """
        return False

    def iterFiles(self):
        """ Iterate through the files matched with the pattern.
        Provide the fileName and fileId.
        """
        filePaths = self.getMatchFiles()

        for fileName in filePaths:
            if self._idRegex:
                # Try to match the file id from filename
                # this is set by the user by using #### format in the pattern
                match = self._idRegex.match(fileName)
                if match is None:
                    raise Exception("File '%s' doesn't match the pattern '%s'"
                                    % (fileName, self.getPattern()))

                fileId = int(match.group(1))

            else:
                fileId = None

            yield fileName, fileId

    @classmethod
    def worksInStreaming(cls):
        # Import protocols always work in streaming
        return True





# **************************************************************************
# *
# * Authors:     Pablo Conesa (pconesa@cnb.csic.es)
# *              J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
""" This module is for Actual wizards to be able to be discovered from the
Domain. wizard.py is left for wizard models and base classes."""
import decimal
import os, json
import requests

import pyworkflow.object as pwobj
from pyworkflow.gui import dialog
import pyworkflow.wizard as pwizard
from pyworkflow.gui.tree import ListTreeProviderString
from pyworkflow.gui.dialog import showError

import pwem.convert as emconv

from pwem.wizards.wizard import EmWizard, FormulaDialog, ColorScaleWizardBase, VariableWizard
import pwem.protocols as emprot
import pwem.viewers as emview
import pwem.objects as emobj
from pwem.constants import RESIDUES3TO1, RESIDUES1TO3


class ImportAcquisitionWizard(EmWizard):
    _targets = [(emprot.ProtImportImages, ['acquisitionWizard'])]

    def show(self, form, *params):
        try:
            acquisitionInfo = form.protocol.loadAcquisitionInfo()
            if isinstance(acquisitionInfo, dict):
                # If acquisitionInfo is None means something is wrong.
                # Now, let's try to show a meaningful error message.
                self._setAcquisition(form, acquisitionInfo)
            else:
                # If not dict, it should be an error message
                dialog.showError("Input error", acquisitionInfo, form.root)

        except FileNotFoundError as e:
            dialog.showInfo("File not found", "Metadata file with acquisition not found.\n\n %s" % e, form.root)

    @classmethod
    def _setAcquisition(cls, form, acquisitionInfo):
        """ Ask whether to set the AcquisitionInfo to the protocol parameters.
        Params:
            acquisitionInfo: Should be a dictionary with acquisition values.
                If None, show an error.
        """
        msg = ''
        for k, v in acquisitionInfo.items():
            msg += '%s = %s\n' % (k, v)
        msg += '\n*Do you want to use detected acquisition values?*'
        response = dialog.askYesNo("Import acquisition",
                                   msg, form.root)
        if response:
            prot = form.protocol
            comment = ''

            for k, v in acquisitionInfo.items():
                if prot.hasAttribute(k):
                    form.setVar(k, v)
                else:
                    comment += "%s = %s\n" % (k, v)
            if comment:
                prot.setObjComment(comment)


class ImportCoordinatesBoxSizeWizard(pwizard.Wizard):
    _targets = [(emprot.ProtImportCoordinates, ['boxSize']),
                (emprot.ProtImportCoordinatesPairs, ['boxSize'])]

    @classmethod
    def _getBoxSize(cls, protocol):
        return protocol.getDefaultBoxSize()

    @classmethod
    def show(cls, form, *params):
        form.setVar('boxSize', cls._getBoxSize(form.protocol))


class ImportOriginVolumeWizard(pwizard.Wizard):
    _targets = [(emprot.ProtImportVolumes, ['x', 'y', 'z'])]

    def show(self, form, *params):
        protocol = form.protocol
        filesPath = protocol.filesPath.get()
        filesPattern = protocol.filesPattern.get()
        if filesPattern:
            fullPattern = os.path.join(filesPath, filesPattern)
        else:
            fullPattern = filesPath

        sampling = protocol.samplingRate.get()
        for fileName, fileId in protocol.iterFiles():
            inputVol = emobj.Volume()
            inputVol.setFileName(fileName)
            if ((str(fullPattern)).endswith('mrc') or
                    (str(fullPattern)).endswith('map')):
                ccp4header = emconv.Ccp4Header(fileName, readHeader=True)
                x, y, z = ccp4header.getOrigin(changeSign=True)  # In Angstroms
            else:
                x, y, z = self._halfOriginCoordinates(inputVol, sampling)

            form.setVar('x', x)
            form.setVar('y', y)
            form.setVar('z', z)

    @classmethod
    def _halfOriginCoordinates(cls, volume, sampling):
        xdim, ydim, zdim = volume.getDim()
        if zdim > 1:
            zdim = zdim / 2.
        x = xdim / 2. * sampling
        y = ydim / 2. * sampling
        z = zdim * sampling
        return x, y, z


class ChangeOriginSamplingWizard(pwizard.Wizard):
    _targets = [(emprot.ProtOrigSampling, ['x', 'y', 'z', 'samplingRate'])]

    def show(self, form, *params):
        protocol = form.protocol
        vol = protocol.inVolume.get()
        fullPattern = vol.getFileName()
        sampling = vol.getSamplingRate()
        if ((str(fullPattern)).endswith('mrc') or
                (str(fullPattern)).endswith('map')):
            ccp4header = emconv.Ccp4Header(fullPattern, readHeader=True)
            x, y, z = ccp4header.getOrigin(changeSign=True)  # In Angstroms
        else:
            x, y, z = \
                ImportOriginVolumeWizard._halfOriginCoordinates(vol, sampling)

        form.setVar('x', round(x, 3))
        form.setVar('y', round(y, 3))
        form.setVar('z', round(z, 3))
        form.setVar('samplingRate', round(sampling, 3))


class GetStructureChainsWizard(pwizard.Wizard):
  # WARNING: this wizard is deprecated. Instead, SelectChainWizard must be used, where inputs, targets and outputs
  # can be specified.
  """Load an atomic structure, parse chain related information as
     name, number of residues, list of aminoacids (or other residues)"""
  _targets = [(emprot.ProtImportSequence, ['inputStructureChain'])
              # NOTE: be careful if you change this class since
              # chimera-wizard inherits from it.
              # (ChimeraModelFromTemplate, ['inputStructureChain'])
              # (atomstructutils, ['inputStructureChain'])
              ]

  @classmethod
  def getModelsChainsStep(cls, protocol):
    """ Returns (1) list with the information
       {"model": %d, "chain": "%s", "residues": %d} (modelsLength)
       (2) list with residues, position and chain (modelsFirstResidue)"""
    structureHandler = emconv.AtomicStructHandler()
    fileName = ""
    if hasattr(protocol, 'pdbId'):
      if protocol.pdbId.get() is not None:
        pdbID = protocol.pdbId.get()
        url = "https://www.rcsb.org/structure/"
        URL = url + ("%s" % pdbID)
        try:
          response = requests.get(URL)
        except:
          raise Exception("Cannot connect to PDB server")
        if (response.status_code >= 400) and (response.status_code < 500):
          raise Exception("%s is a wrong PDB ID" % pdbID)
        fileName = structureHandler.readFromPDBDatabase(
          os.path.basename(pdbID), dir="/tmp/")
      else:
        fileName = protocol.pdbFile.get()
    else:
      if protocol.pdbFileToBeRefined.get() is not None:
        fileName = os.path.abspath(protocol.pdbFileToBeRefined.get(
        ).getFileName())

    structureHandler.read(fileName)
    structureHandler.getStructure()
    # listOfChains, listOfResidues = structureHandler.getModelsChains()
    return structureHandler.getModelsChains()

  def editionListOfChains(self, listOfChains):
    self.chainList = []
    for model, chainDic in listOfChains.items():
      for chainID, lenResidues in chainDic.items():
        self.chainList.append(
          '{"model": %d, "chain": "%s", "residues": %d}' %
          (model, str(chainID), lenResidues))

  def show(self, form, *params):
    protocol = form.protocol
    try:
      listOfChains, listOfResidues = self.getModelsChainsStep(protocol)
    except Exception as e:
      print("ERROR: ", e)
      return

    self.editionListOfChains(listOfChains)
    finalChainList = []
    for i in self.chainList:
      finalChainList.append(pwobj.String(i))
    provider = ListTreeProviderString(finalChainList)
    dlg = dialog.ListDialog(form.root, "Model chains", provider,
                            "Select one of the chains (model, chain, "
                            "number of chain residues)")
    form.setVar('inputStructureChain', dlg.values[0].get())


class SelectChainWizard(VariableWizard):
    '''Opens the input AtomStruct and allows you to select one of the present chains'''
    _targets, _inputs, _outputs = [], {}, {}

    @classmethod
    def getModelsChainsStep(cls, protocol, inputObj):
        """ Returns (1) list with the information
           {"model": %d, "chain": "%s", "residues": %d} (modelsLength)
           (2) list with residues, position and chain (modelsFirstResidue)"""
        structureHandler = emconv.AtomicStructHandler()
        if type(inputObj) == str:
            if os.path.exists(inputObj):
                fileName = inputObj
            else:
                pdbID = inputObj
                url = "https://www.rcsb.org/structure/"
                URL = url + ("%s" % pdbID)
                try:
                    response = requests.get(URL)
                except:
                    raise Exception("Cannot connect to PDB server")
                if (response.status_code >= 400) and (response.status_code < 500):
                    raise Exception("%s is a wrong PDB ID" % pdbID)
                fileName = structureHandler.readFromPDBDatabase(os.path.basename(pdbID), dir="/tmp/")

        elif str(type(inputObj).__name__) == 'SchrodingerAtomStruct':
            fileName = os.path.abspath(inputObj.convert2PDB())
        else:
            fileName = os.path.abspath(inputObj.getFileName())

        structureHandler.read(fileName)
        structureHandler.getStructure()
        return structureHandler.getModelsChains()

    def editionListOfChains(self, listOfChains):
        chainList = []
        for model, chainDic in listOfChains.items():
            for chainID, lenResidues in chainDic.items():
                chainList.append(
                    '{"model": %d, "chain": "%s", "residues": %d}' %
                    (model, str(chainID), lenResidues))
        return chainList

    def show(self, form, *params):
        inputParams, outputParam = self.getInputOutput(form)
        protocol = form.protocol
        try:
            inputObj = getattr(protocol, inputParams[0]).get()
            listOfChains, listOfResidues = self.getModelsChainsStep(protocol, inputObj)
        except Exception as e:
            print("ERROR: ", e)
            return

        chainList = self.editionListOfChains(listOfChains)
        finalChainList = []
        for i in chainList:
            finalChainList.append(pwobj.String(i))
        provider = ListTreeProviderString(finalChainList)
        dlg = dialog.ListDialog(form.root, "Model chains", provider,
                                "Select one of the chains (model, chain, "
                                "number of chain residues)")
        form.setVar(outputParam[0], dlg.values[0].get())

SelectChainWizard().addTarget(protocol=emprot.ProtImportSequence,
                              targets=['inputStructureChain'],
                              inputs=[['pdbId', 'pdbFile']],
                              outputs=['inputStructureChain'])


class SelectResidueWizard(SelectChainWizard):
    _targets, _inputs, _outputs = [], {}, {}

    def editionListOfResidues(self, modelsFirstResidue, model, chain):
      residueList = []
      for modelID, chainDic in modelsFirstResidue.items():
        if int(model) == modelID:
          for chainID, seq_number in chainDic.items():
            if chain == chainID:
              for i in seq_number:
                residueList.append('{"index": %d, "residue": "%s"}' % (i[0], str(i[1])))
      return residueList

    def getResidues(self, form, inputObj, chainStr):
      protocol = form.protocol

      if type(inputObj) == str:
          # Select the residues if the input structure parameter is a str (PDB id or PDB file)
          structureHandler = emconv.AtomicStructHandler()
          if os.path.exists(inputObj):
              fileName = inputObj
          else:
              pdbID = inputObj
              url = "https://www.rcsb.org/structure/"
              URL = url + ("%s" % pdbID)
              try:
                response = requests.get(URL)
              except:
                raise Exception("Cannot connect to PDB server")
              if (response.status_code >= 400) and (response.status_code < 500):
                raise Exception("%s is a wrong PDB ID" % pdbID)

              fileName = structureHandler.readFromPDBDatabase(os.path.basename(pdbID), dir="/tmp/")

          structureHandler.read(fileName)
          structureHandler.getStructure()
          modelsLength, modelsFirstResidue = structureHandler.getModelsChains()

          struct = json.loads(chainStr)  # From wizard dictionary
          chain, model = struct["chain"].upper().strip(), int(struct["model"])

          residueList = self.editionListOfResidues(modelsFirstResidue, model, chain)
          finalResiduesList = []
          for i in residueList:
              finalResiduesList.append(emobj.String(i))

      elif issubclass(type(inputObj), emobj.AtomStruct):
          try:
            modelsLength, modelsFirstResidue = self.getModelsChainsStep(protocol, inputObj)
          except Exception as e:
            print("ERROR: ", e)
            return
          struct = json.loads(chainStr)  # From wizard dictionary
          chain, model = struct["chain"].upper().strip(), int(struct["model"])

          residueList = self.editionListOfResidues(modelsFirstResidue, model, chain)
          finalResiduesList = []
          for i in residueList:
            finalResiduesList.append(emobj.String(i))

      elif issubclass(type(inputObj), emobj.Sequence) or str(type(inputObj).__name__) == 'SequenceVariants':
          finalResiduesList = []
          for i, res in enumerate(inputObj.getSequence()):
            if res in RESIDUES1TO3:
                res3 = RESIDUES1TO3[res]
            else:
                res3 = res
            stri = '{"index": %s, "residue": "%s"}' % (i + 1, res3)
            finalResiduesList.append(emobj.String(stri))

      return finalResiduesList

    def getSequence(self, finalResiduesList, idxs):
        roiStr, inSeq = '', False
        for residue in finalResiduesList:
            resDic = json.loads(residue.get())
            if resDic['residue'] in RESIDUES3TO1:
                if resDic['index'] == idxs[0]:
                  inSeq = True
                if resDic['index'] == idxs[-1]:
                  inSeq = False
                  roiStr += RESIDUES3TO1[resDic['residue']]
                  break

                if inSeq:
                    roiStr += RESIDUES3TO1[resDic['residue']]
        return roiStr

    def show(self, form, *params):
      inputParams, outputParam = self.getInputOutput(form)
      protocol = form.protocol
      inputObj = getattr(protocol, inputParams[0]).get()
      if len(inputParams) < 2:
          # For sequence objects, with no chain
          chainStr = None
      else:
          chainStr = getattr(protocol, inputParams[1]).get()
      finalResiduesList = self.getResidues(form, inputObj, chainStr)

      provider = ListTreeProviderString(finalResiduesList)
      dlg = dialog.ListDialog(form.root, "Chain residues", provider,
                              "Select one residue (residue number, "
                              "residue name)")

      idxs = [json.loads(dlg.values[0].get())['index'], json.loads(dlg.values[-1].get())['index']]
      roiStr = self.getSequence(finalResiduesList, idxs)


      intervalStr = '{"index": "%s-%s", "residues": "%s"}' % (idxs[0], idxs[1], roiStr)
      form.setVar(outputParam[0], intervalStr)


class PythonFormulaeWizard(pwizard.Wizard):
    """Assist in the creation of python formula to be evaluated. In Steps"""
    _targets = [(emprot.ProtSetFilter, ['formula']),
                (emprot.ProtSetEditor, ['formula'])]

    def show(self, form, *params):

        d = FormulaDialog(form.root, form.protocol.inputSet.get(), formula=form.protocol.formula.get())

        # If accepted
        if d.resultYes():
            form.setVar('formula',d.getFormula())


class PythonFormulaWizardX1(pwizard.Wizard):
    """Assist in the creation of python formula to be evaluated. In Steps"""
    _targets = [(emprot.ProtMathematicalOperator, ['attribute1'])]

    def show(self, form, *params):

        d = FormulaDialog(form.root, form.protocol.inputSet1.get(), formula=form.protocol.attribute1.get())

        # If accepted
        if d.resultYes():
            form.setVar('attribute1', d.getFormula())


class PythonFormulaWizardX2(pwizard.Wizard):
    """Assist in the creation of python formula to be evaluated. In Steps"""
    _targets = [(emprot.ProtMathematicalOperator, ['attribute2'])]

    def show(self, form, *params):

        d2 = FormulaDialog(form.root, form.protocol.inputSet2.get(), formula=form.protocol.attribute2.get())

        # If accepted
        if d2.resultYes():
            form.setVar('attribute2', d2.getFormula())


class PythonTopRankWizard(pwizard.Wizard):
    """Assist in the creation of python formula to be evaluated. In Steps"""
    _targets = [(emprot.ProtSetFilter, ['rankingField'])]

            

class SelectAttributeWizard(VariableWizard):
    """Wizard to select attributes stored in a scipion object or set """
    _targets, _inputs, _outputs = [], {}, {}

    def getInputSet(self, form, inputParam):
      inputPointer = getattr(form.protocol, inputParam)
      if issubclass(inputPointer.__class__, pwobj.PointerList):
        inputPointer = inputPointer[0]
      return inputPointer.get()

    def getFirstItem(self, form, inputParam):
        inputSet = self.getInputSet(form, inputParam)
        if issubclass(inputSet.__class__, pwobj.Set):
            item = inputSet.getFirstItem()
        elif issubclass(inputSet.__class__, pwobj.Object):
            item = inputSet
        return item

    def getInputAttributes(self, form, inputParam):
      attrNames = []
      item = self.getFirstItem(form, inputParam[0])
      for key, attr in item.getAttributesToStore():
        attrNames.append(key)
      return attrNames

    def show(self, form, *params):
      inputParam, outputParam = self.getInputOutput(form)
      if self.getInputSet(form, inputParam[0]):
        attrsList = self.getInputAttributes(form, inputParam)
        finalAttrsList = []
        for i in attrsList:
          finalAttrsList.append(pwobj.String(i))
        provider = ListTreeProviderString(finalAttrsList)
        dlg = dialog.ListDialog(form.root, "Filter set", provider,
                              "Select one of the attributes")
        form.setVar(outputParam[0], dlg.values[0].get())
      else:
        showError('Input error', 'Input needs to be defined in the form for the wizard be used', form.root)

class ColorScaleWizardRMSD(ColorScaleWizardBase):
    _targets = ColorScaleWizardBase.defineTargets(emview.ChimeraAttributeViewer)


#Defining target for the SelectAttributeWizard
SelectAttributeWizard().addTarget(protocol=emprot.ProtSetFilter,
                                  targets=['rankingField'],
                                  inputs=['inputSet'],
                                  outputs=['rankingField'])
